# -*- coding: utf-8 -*-
import ase
import numpy as np

from milad import models
from milad import generate
from milad import fingerprinting


def test_create_fingerprint_set():
    # Settings
    np.random.seed(5)
    num_atoms = 6
    cutoff = 2.5
    descriptor = fingerprinting.descriptor(cutoff=cutoff)

    positions = generate.random_points_in_sphere(num_atoms, cutoff)
    atoms = ase.Atoms(positions=positions, numbers=(1.,) * num_atoms)

    fingerprint_set = models.create_fingerprint_set(descriptor, [atoms], get_derivatives=True)
    assert len(fingerprint_set) == 1



# -*- coding: utf-8 -*-
import numpy as np
import plotly.graph_objects as go


def isosurface(grid: np.ndarray, values: np.ndarray, **kwargs) -> go.Isosurface:
    """Helper for creating a plotly isosurface from an array of grid coordinates and corresponding values"""
    params = dict(isomax=np.inf)
    params.update(kwargs)

    return go.Isosurface(x=grid[:, 0], y=grid[:, 1], z=grid[:, 2], value=values, **params)


def volume(grid: np.ndarray, values: np.ndarray, **kwargs):
    params = dict(surface_count=25)
    params.update(kwargs)

    return go.Volume(x=grid[:, 0], y=grid[:, 1], z=grid[:, 2], value=values, **params)


def scatter3d(points: np.ndarray, **kwargs) -> go.Scatter3d:
    """Helper for creating a plotly 3d scatter plot from an array of points"""
    params = dict(mode='markers')
    params.update(kwargs)
    return go.Scatter3d(x=points[:, 0], y=points[:, 1], z=points[:, 2], **params)



# -*- coding: utf-8 -*-
import collections
import math
from typing import Union, Callable, Sequence, Optional, List

import ase
import torch
import torch.nn.functional
from torch.autograd import functional

from milad.play import asetools
from milad import fingerprinting
from milad import utils

__all__ = 'NeuralNetwork', 'create_fingerprint_set'

# pylint: disable=no-member, not-callable


def create_fingerprint_set(
    descriptor: fingerprinting.MomentInvariantsDescriptor,
    systems: Sequence[ase.Atoms],
    get_derivatives=False
) -> utils.FingerprintSet:
    # WARNING: The calculation of derivatives only takes into account positional degrees of freedom (not species) and
    # makes assumptions about the shape of the derivatives tensor

    fp_length = descriptor.fingerprint_len

    fingerprints = utils.FingerprintSet(descriptor.fingerprint_len)
    for system in systems:
        fps = []
        fp_derivatives = []
        for env in asetools.extract_environments(system, cutoff=descriptor.cutoff):
            milad_env = asetools.ase2milad(env)
            if get_derivatives:
                fingerprint, jacobian = descriptor(milad_env, jacobian=True)
                # Now, let's deal with the derivatives extracting just the positional parts and summing over all
                # neighbours keeping x, y, z separate
                derivatives = jacobian[:, 3:3 * len(env)].reshape(fp_length, -1, 3).sum(1)

                fps.append(fingerprint)
                fp_derivatives.append(derivatives)
            else:
                fps.append(descriptor(milad_env))

        fingerprints.add_system(system, fps, derivatives=fp_derivatives)

    return fingerprints


class Predictions:

    def __init__(self, sizes: torch.Tensor, local_energies: torch.Tensor, forces: torch.Tensor = None):
        """Create an object to store prediction data from the neural network

        :param sizes: a tensor containing the number of atoms in each system.  The length of the tensor is the total
            number of systems the prediction is being made for.
        """
        if not sizes.sum() == len(local_energies):
            raise ValueError(
                f"The total number of atoms ({sizes.sum()}) doesn't match the number of "
                f'local energies ({len(local_energies)}).'
            )

        self.sizes = sizes
        self.local_energies = local_energies
        self.forces = forces
        self._total_energies = None

    @property
    def total_energies(self) -> torch.Tensor:
        """Get total energies as a single tensor that preserves gradients"""
        if self._total_energies is None:
            # Lazily calculate
            idx = 0
            total_energies = []
            for num_atoms in self.sizes:
                total_energies.append(torch.sum(self.local_energies[idx:idx + num_atoms]))
                idx += num_atoms
            # Combine them all into a tensor
            self._total_energies = torch.stack(total_energies)

        return self._total_energies

    def get_normalised_energies(self) -> torch.Tensor:
        """Get the energy/atom for each system"""
        return self.total_energies / self.sizes


class FittingData:
    """Data used during a fitting procedure"""

    @classmethod
    def from_fingerprint_set(cls, fingerprint_set: utils.FingerprintSet, device=None, dtype=None, requires_grad=False):
        tensor_kwargs = dict(dtype=dtype, device=device, requires_grad=requires_grad)
        fingerprints = torch.cat(tuple(torch.tensor(fps, **tensor_kwargs) for fps in fingerprint_set.fingerprints))

        forces = None
        if fingerprint_set.has_all_forces():
            forces = torch.cat(
                tuple(
                    torch.tensor(forces, **tensor_kwargs) if forces is not None else None
                    for forces in fingerprint_set.get_forces()
                )
            )

        derivatives = None
        if fingerprint_set.has_all_derivatives():
            derivatives = torch.cat(
                tuple(
                    torch.tensor(derivatives, **tensor_kwargs) if derivatives is not None else None
                    for derivatives in fingerprint_set.fingerprint_derivatives
                )
            )

        return FittingData(
            torch.tensor(fingerprint_set.sizes, dtype=torch.int, device=device),
            fingerprints,
            total_energies=torch.tensor(fingerprint_set.get_potential_energies(normalise=False), **tensor_kwargs),
            forces=forces,
            derivatives=derivatives
        )

    def __init__(
        self,
        sizes: torch.Tensor,
        fingerprints: torch.Tensor,
        total_energies: torch.Tensor,
        forces: Optional[torch.Tensor] = None,
        derivatives: Optional[torch.Tensor] = None
    ):
        """
        Construct a set of fitting data.

        :param fingerprints: a tuple containing a set of fingerprints for each atom in the structure
        :param total_energies: the corresponding energy for each structure
        :param forces: the (optional) forces for each atom in the structure
        """
        if sizes.sum() != len(fingerprints):
            raise ValueError(
                f"The total number of atoms ({sizes.sum()}) doesn't match the number of "
                f'fingerprints ({len(fingerprints)}).'
            )
        if len(sizes) != len(total_energies):
            raise ValueError(
                f"The number of total energies ({len(total_energies)}) doesn't match the number of "
                f'systems (sizes) ({len(sizes)}).'
            )

        if not (len(derivatives) == len(fingerprints) == len(forces)):  # pylint: disable=superfluous-parens
            raise ValueError('Dataset size mismatch')

        self._num_atoms = sizes
        self._fingerprints = fingerprints
        self._total_energies = total_energies
        self._forces = forces
        self._derivatives = derivatives

    def __getitem__(self, item) -> 'FittingData':
        if isinstance(item, slice):
            atomic_slice = self._get_per_atom_slice(item)

            sizes = self._num_atoms[item]
            total_energies = self._total_energies[item]
            fingerprints = self._fingerprints[atomic_slice]
            forces = self._forces[atomic_slice] if self._forces is not None else None
            derivatives = self._derivatives[atomic_slice] if self._derivatives is not None else None

            return FittingData(sizes, fingerprints, total_energies, forces, derivatives)

        # Assume item is a fixed index
        return self.__getitem__(slice(item, item + 1))

    def __len__(self) -> int:
        """Returns the total number of systems contained in this fitting data"""
        return len(self._num_atoms)

    def _get_per_atom_slice(self, item: slice) -> slice:
        if item.step is not None and item.step != 1:
            raise ValueError(f'Non-contigious slices are unsupported for now, got step {item.step}')

        start = self._num_atoms[:item.start].sum().cpu().item()
        stop = self._num_atoms[:item.stop].sum().cpu().item()
        return slice(start, stop)

    @property
    def fingerprints(self) -> torch.Tensor:
        """Fingerprints tensors.  One entry per system."""
        return self._fingerprints

    @property
    def total_energies(self) -> torch.Tensor:
        """Get the known energies"""
        return self._total_energies

    @property
    def forces(self) -> Optional[torch.Tensor]:
        """Get the known forces if we have them"""
        return self._forces

    @property
    def derivatives(self) -> Optional[torch.Tensor]:
        return self._derivatives

    @property
    def num_atoms(self) -> torch.Tensor:
        """Get the number of atoms in each structure"""
        return self._num_atoms

    def get_normalised_energies(self) -> torch.Tensor:
        return self._total_energies / self._num_atoms

    def batch_split(self, batchsize: int) -> List['FittingData']:
        total_samples = len(self)
        num_batches = int(math.ceil(total_samples / batchsize))
        batches = []
        for batch_num in range(num_batches):
            start_idx = batch_num * batchsize
            end_idx = min(start_idx + batchsize, total_samples)
            batches.append(self[start_idx:end_idx])

        return batches


class LossFunction:
    Loss = collections.namedtuple('Result', 'energy force total')

    def __init__(self, energy_coeff=1., force_coeff=0.1):
        self.energy_coeff = energy_coeff
        self.force_coeff = force_coeff

    def get_loss(self, predictions: Predictions, fitting_data: FittingData):
        total_loss = 0.
        energy_loss = torch.nn.functional.mse_loss(
            predictions.get_normalised_energies(), fitting_data.get_normalised_energies()
        )
        total_loss += self.energy_coeff * energy_loss

        force_loss = None
        if fitting_data.forces is not None and self.force_coeff != 0.:
            force_loss = torch.nn.functional.mse_loss(predictions.forces, fitting_data.forces) / 3.
            total_loss += self.force_coeff * force_loss
            total_loss *= 0.5

        return LossFunction.Loss(energy_loss, force_loss, total_loss)


class Range:
    """A range object representing a minimum and maximum value"""

    def __init__(self, initial=(0., 1.), device=None, dtype=None):
        initial = initial or (0., 1.)
        self._range = torch.tensor(initial, device=device, dtype=dtype)
        self._using_defaults = True

    def __str__(self):
        return '[{}, {}]'.format(self.min.item(), self.max.item())

    @property
    def range(self):
        return self._range

    @property
    def min(self):
        return self._range[0]

    @property
    def max(self):
        return self._range[1]

    @property
    def span(self):
        return self.max - self.min

    def expand(self, values: torch.Tensor):
        """Expand the range to encompass the passed values"""
        if self._using_defaults:
            self._range[0] = values.min()
            self._range[1] = values.max()
            self._using_defaults = False
        else:
            vals_min = values.min()
            vals_max = values.max()
            self._range[0] = self.min if self.min < vals_min else vals_min
            self._range[1] = self.max if self.max > vals_max else vals_max


class DataScaler(torch.nn.Module):
    """Scale data to be in the range in a particular range.  Handles scaling on both input and output sides
    so if you can set the range of input and outputs separately."""

    def __init__(self, in_range=(0., 1.), out_range=(0., 1.), device=None, dtype=None):
        super().__init__()
        # Default to range [-1, 1]
        self._in_range = Range(in_range, device=device, dtype=dtype)  # Input range
        self._out_range = Range(out_range, device=device, dtype=dtype)  # Output range

    def __str__(self):
        return '{} -> {}'.format(self._in_range, self._out_range)

    @property
    def input(self) -> Range:
        return self._in_range

    @property
    def output(self) -> Range:
        return self._out_range

    def scale(self, x: torch.Tensor):  # pylint: disable=invalid-name
        return self.output.span * (x - self.input.min) / self.input.span + self.output.min

    def unscale(self, y: torch.Tensor):  # pylint: disable=invalid-name
        return (y - self.output.min) * self.input.span / self.output.span + self.input.min

    def forward(self, x: torch.Tensor):  # pylint: disable=invalid-name
        return self.scale(x)


class NeuralNetwork:
    """Neural network empirical potential model"""

    def __init__(
        self,
        hiddenlayers=(32, 16, 8),
        activations: Union[str, Callable] = 'ReLU',
        loss_function=None,
        device=None,
        dtype=torch.float64,
        bias=False,
        vectorise_jacobian=False,
    ):
        """

        :param hiddenlayers: a tuple of the hidden layers to use e.g. (32, 16, 8)
        :param activations:
        :param loss_function:
        :param device:
        :param dtype:
        :param bias:
        :param vectorise_jacobian: if True, will calculate the NN Jacobian for a batch of fingerprints all in one go.
            This is very memory consuming at the moment and may exceed the amount of memory you have available.
        """
        # Set up the hidden layers
        self._hidden = hiddenlayers

        # Set the activation function
        if isinstance(activations, str):
            self._activations = getattr(torch.nn, activations)()
        elif isinstance(activations, Callable):  # pylint: disable=isinstance-second-argument-not-valid-type
            self._activations = activations
        else:
            raise TypeError('Expecting str or Callable, got {}'.format(activations.__class__.__name__))

        if device is None:
            self._device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        else:
            self._device = device
        self._dtype = dtype
        self._bias = bias
        self._vectorise_jacobian = vectorise_jacobian

        self._fingerprint_scaler = DataScaler(out_range=(-1., 1.), device=self._device, dtype=self._dtype)
        self._energy_scaler = DataScaler(in_range=(-1., 1.), device=self._device, dtype=self._dtype)
        self.loss_function = loss_function or LossFunction()

        self._network = None

    def fit(
        self,
        training_set: Union[Sequence[ase.Atoms], utils.FingerprintSet],
        max_epochs=200,
        progress_callback: Callable = None,
        learning_rate=5e-4,
        batchsize=16,
    ):
        if self._network is None:
            self._create_network(training_set.fingerprint_len)

        training_data = self.create_fitting_data(training_set, requires_grad=True)

        # Scale the inputs and outputs to match the ranges
        norm_energies = training_data.get_normalised_energies()
        self._energy_scaler.output.expand(norm_energies)

        def stopping_function(epoch, _training, _loss):
            if epoch >= max_epochs:
                return True

            return False  # Don't stop

        optimiser = torch.optim.Adam(self._network.parameters(), lr=learning_rate)
        loss = self._train(
            training_data,
            stopping_function,
            optimiser,
            batchsize,
            progress_callback=progress_callback,
        )
        return training_data, loss

    def create_fitting_data(self, fingerprint_set: utils.FingerprintSet, requires_grad=False):
        return FittingData.from_fingerprint_set(
            fingerprint_set,
            device=self._device,
            dtype=self._dtype,
            requires_grad=requires_grad,
        )

    def _train(
        self,
        training: FittingData,
        should_stop: Callable,
        optimiser,
        batchsize: int,
        progress_callback: Callable = None,
    ) -> LossFunction.Loss:
        # Let's break up the training data into batches
        batches = training.batch_split(batchsize)

        epoch = 0
        loss_result = None
        get_forces = self.loss_function.force_coeff != 0.
        while True:
            for batch in batches:
                predictions = self.make_prediction(batch, get_forces=get_forces, create_graph=True)

                optimiser.zero_grad()
                loss_result = self.loss_function.get_loss(predictions, batch)
                loss = loss_result.total
                loss.backward(retain_graph=True)
                optimiser.step()

            # Calculate loss for the entire training set.  Only necessary if there is more than one batch
            if len(batches) > 1:
                predictions = self.make_prediction(training, get_forces=get_forces)
                loss_result = self.loss_function.get_loss(predictions, training)

            if progress_callback is not None:
                progress_callback(self, epoch, training, loss_result)

            if should_stop(epoch, training, loss_result):
                break

            epoch += 1

        return loss_result

    def loss(self, fitting_data: FittingData, get_forces=None) -> LossFunction.Loss:
        get_forces = get_forces if get_forces is not None else self.loss_function.force_coeff != 0.
        predictions = self.make_prediction(fitting_data, get_forces=get_forces)
        return self.loss_function.get_loss(predictions, fitting_data)

    def make_prediction(self, fitting_data: FittingData, get_forces=False, create_graph=False) -> Predictions:
        # Join all fingerprints as this is much faster going through the network
        local_energies = self._network(fitting_data.fingerprints)

        forces = None
        if get_forces:
            forces = []
            # Need to calculate the forces using the chain rule with the fingerprints derivative and the
            # neural network derivative
            if self._vectorise_jacobian:
                jac = functional.jacobian(
                    self._network, fitting_data.fingerprints, create_graph=create_graph, vectorize=True
                )
                jacs = [jac[i, :, i, :] for i in range(len(fitting_data.fingerprints))]

                for fingerprint, derivatives, network_deriv in zip(
                    fitting_data.fingerprints, fitting_data.derivatives, jacs
                ):
                    # Get the derivative of the energy wrt to input vector
                    predicted_force = -torch.matmul(network_deriv, derivatives)
                    forces.append(predicted_force)
            else:
                # Fallback, lower memory version
                for fingerprint, derivatives in zip(fitting_data.fingerprints, fitting_data.derivatives):
                    # Get the derivative of the energy wrt to input vector
                    network_deriv = functional.jacobian(
                        self._network, fingerprint, create_graph=create_graph, vectorize=True
                    )
                    predicted_force = -torch.matmul(network_deriv, derivatives)
                    forces.append(predicted_force)

            # VMAP version, doesn't work yet but may be fixed soon and would almost certainly be fast and relatively
            # low-memory.  See:
            # https://github.com/pytorch/pytorch/issues/42368
            # def get_jacobian(fingerprint):
            #     return functional.jacobian(self._network, fingerprint, create_graph=False)
            #
            # jac = torch.vmap(get_jacobian)(tuple(fitting_data.fingerprints))
            # jacs = [jac[i, :, i, :] for i in range(len(fitting_data.fingerprints))]
            #
            # for fingerprint, derivatives, network_deriv
            #   in zip(fitting_data.fingerprints, fitting_data.derivatives, jacs):
            #     # Get the derivative of the energy wrt to input vector
            #     predicted_force = -torch.matmul(network_deriv, derivatives)
            #     forces.append(predicted_force)

            forces = torch.cat(forces)

        return Predictions(fitting_data.num_atoms, local_energies=local_energies, forces=forces)

    def _create_network(self, input_size):
        # Create the network
        # sequence = [self._fingerprint_scaler]
        sequence = []
        prev_size = input_size
        # Create the fully connected (hidden) layers
        for size in self._hidden:
            sequence.append(torch.nn.Linear(prev_size, size, bias=self._bias))
            sequence.append(self._activations)
            prev_size = size
        # Add the output layer
        sequence.extend((torch.nn.Linear(prev_size, 1, bias=self._bias), self._energy_scaler))

        self._network = torch.nn.Sequential(*sequence)
        self._network.to(self._device, self._dtype)

    def tensor(self, *args, requires_grad=False) -> torch.Tensor:
        return torch.tensor(*args, device=self._device, dtype=self._dtype, requires_grad=requires_grad)



# -*- coding: utf-8 -*-
import numpy as np


def random_points_in_sphere(num: int, radius=1., centre=False) -> np.array:
    """Generate num points within a sphere of the given radius"""
    points = np.empty((num, 3))
    r_sq = radius * radius

    for idx in range(num):
        vec = 2 * radius * (np.random.rand(3) - 0.5)
        while np.dot(vec, vec) >= r_sq:
            vec = 2 * radius * (np.random.rand(3) - 0.5)
        points[idx] = vec

    if centre:
        points -= centeroid(points)
        # Make sure that there aren't any outside of the radius now
        radius_sq = (points**2).sum(1).max()
        if radius_sq > r_sq:
            points *= r_sq / radius_sq

    return points


def centeroid(pts: np.ndarray) -> np.ndarray:
    length = len(pts)
    centroid = np.empty((length, 3))
    centroid[:, 0] = np.sum(pts[:, 0]) / length
    centroid[:, 1] = np.sum(pts[:, 1]) / length
    centroid[:, 2] = np.sum(pts[:, 2]) / length
    return centroid



# -*- coding: utf-8 -*-
import numpy as np
import pytest

from milad import atomic


def test_map_numbers():
    """Test that the mapping atomic numbers function works correctly"""
    num_atoms = 6
    system = atomic.AtomsCollection(
        num_atoms, positions=np.random.rand(num_atoms, 3), numbers=np.array([1, 2, 5, 2, 3, 8])
    )

    # Map the five species into the range 2 -> 3
    mapper = atomic.MapNumbers(set(system.numbers), map_to=(2., 3.))
    mapped = mapper(system)
    assert mapped.numbers[0] == pytest.approx(2.1)
    assert mapped.numbers[1] == pytest.approx(2.3)
    assert mapped.numbers[2] == pytest.approx(2.7)
    assert mapped.numbers[3] == pytest.approx(2.3)
    assert mapped.numbers[4] == pytest.approx(2.5)
    assert mapped.numbers[5] == pytest.approx(2.9)

    # Now test inversion
    inverted = mapper.inverse(mapped)
    diffs = inverted.numbers - system.numbers
    assert diffs.max() == pytest.approx(0.)

    # Now check that we can also map all species down to a single number (i.e. discard species info)
    mapper = atomic.MapNumbers(set(system.numbers), map_to=3.4)
    mapped = mapper(system)
    assert np.all(mapped.numbers == 3.4)


def test_separation_force():
    """Test that the separation force works correctly"""
    separation_force = atomic.SeparationForce(force_constant=1, cutoff=1.)
    separation = 0.5
    atoms = atomic.AtomsCollection(2, positions=[[-0.5 * separation, 0, 0], [0.5 * separation, 0, 0]], numbers=1.)

    energy, jacobian = separation_force(atoms, jacobian=True)
    assert energy == separation_force.energy(separation)
    assert energy > 0., 'The energy should always be positive if the separation less than the cutoff'

    force = separation_force.force(separation)
    assert force > 0., 'The force should act to separate the two atoms'

    assert np.all(jacobian[0, atoms.linear_pos_idx(0)] == np.array([force, 0., 0.]))
    assert np.all(jacobian[0, atoms.linear_pos_idx(1)] == np.array([-force, 0., 0.]))

    # Check that the cutoff works
    atoms.positions[1, 0] = 1.
    energy, jacobian = separation_force(atoms, jacobian=True)
    assert energy == 0.0
    assert np.all(jacobian == 0.)



# -*- coding: utf-8 -*-
import sympy
from sympy.physics.quantum.cg import CG

from milad import utils

# We're using a lot of variable names that make sense from the mathematics but are not so programming friendly
# so disable for now as we know what they mean
# pylint: disable=invalid-name


def composite_complex_moment_form(moments: sympy.Indexed, n: int, l1: int, l2: int, j: int, k: int):
    """Get the composite moment form as defined by Lo and Don denoted as:

    c_n(l, l')_j^k
    """
    if not l1 >= 0:
        raise ValueError(f'l must be greater than 0, got {l1}')
    if not l2 >= 0:
        raise ValueError(f'l2 must be greater than 0, got {l2}')
    if not utils.even(j):
        raise ValueError('j must be even, got {}'.format(j))
    if not utils.even(n - l1):
        raise ValueError(f'n - l must be even got {n} - {l1}')
    if not utils.even(n - l2):
        raise ValueError(f'n - l2 must be even got {n} - {l2}')
    eqn = None

    for m in utils.inclusive(max(-l1, k - l2), min(l1, k + l2), 1):
        term = CG(l1, m, l2, k - m, j, k).doit() * moments[n, l1, m] * moments[n, l2, k - m]
        eqn = term if eqn is None else eqn + term

    assert eqn is not None, (max(-l1, k - l2), min(l1, k + l2))
    return eqn


def moment_form_complex_moment(moments: sympy.Indexed, n1: int, l1: int, l2: int, n2: int, j: int):
    """Get a moment form combined with complex moments i.e.:

    c_s(l, l')_j c_{n'}
    """
    if not j >= 0:
        raise ValueError(f"'j' must be >= 0, got {j}")
    if not utils.even(n2 - j):
        raise ValueError(f'n2 - j must be even got {n2} - {j}')

    eqn = None

    for k in utils.inclusive(-j, j, 1):
        term = (-1)**(j - k) * composite_complex_moment_form(moments, n1, l1, l2, j, k) * moments[n2, j, -k]
        eqn = term if eqn is None else eqn + term

    return 1 / sympy.sqrt(2 * j + 1) * eqn


def moment_form_moment_form(moments: sympy.Indexed, n1: int, l1: int, l2: int, n2: int, l3: int, l4: int, j: int):
    """Get a complex moment form combined with a complex moment form i.e.:

    c_n(l, l')_j c_{n'}(l'', l''')_j
    """
    eqn = None

    for k in utils.inclusive(-j, j, 1):
        term = (-1) ** (j - k) * \
               composite_complex_moment_form(moments, n1, l1, l2, j, k) * \
               composite_complex_moment_form(moments, n2, l3, l4, j, -k)
        eqn = term if eqn is None else eqn + term

    return 1 / sympy.sqrt(2 * j + 1) * eqn


def invariants_single_complex_moments(moments: sympy.Indexed, max_n: int):
    """Get invariants from single complex moments i.e.:

    c_{n0}^0
    """
    invariants = []
    for n in utils.inclusive(0, max_n, 2):
        invariants.append(moments[n, 0, 0])
    return invariants


def invariants_single_complex_moment_forms(moments: sympy.Indexed, max_n: int):
    """Get invariants from single complex moment forms i.e.:

    c_n(l, l)_0^0

    n = 1, 2, .., max_n
    l = n, n - 2, n - 4, ..., 1
    """
    invariants = []
    for n in utils.inclusive(1, max_n, 1):
        for l in utils.inclusive(n, 1, -2):
            invariant = composite_complex_moment_form(moments, n=n, l1=l, l2=l, j=0, k=0)
            invariants.append(invariant)
    return invariants


def invariants_moment_form_complex_moment(moments: sympy.Indexed, max_n: int):
    """Get invariants from a complex moment form and a complex moment i.e.:


    c_n(l, l')_j c_{n'}
    """
    invariants = []
    for n in utils.inclusive(1, max_n, 1):
        for n2 in utils.inclusive(2, n, 2):
            for j in utils.inclusive(2, n2, 2):
                for l1 in utils.inclusive(n, 1, -2):
                    for l2 in utils.inclusive(l1, max(l1 - j, 1), -2):
                        assert l2 >= l1 - j
                        assert l2 >= 1
                        invariant = moment_form_complex_moment(moments, n1=n, l1=l1, l2=l2, n2=n2, j=j)
                        invariants.append(invariant)

    return invariants


def invariants_two_moment_forms(moments: sympy.Indexed, max_n: int, verbose=False):
    """Get invariants from two complex moment forms i.e.:


    c_n(l, l')_j c_{n'}(l'', l''')_j
    """
    invariants = []
    for n in utils.inclusive(1, max_n, 1):
        for n_prime in utils.inclusive(1, n, 1):
            for j in utils.inclusive(0, n_prime, 2):
                for l in utils.inclusive(n, 1, -2):
                    for l2 in utils.inclusive(l, max(l - j, 1), -2):
                        for l3 in utils.inclusive(n_prime, 1, -2):
                            for l4 in utils.inclusive(l3, max(l3 - j, 1), -2):
                                invariant = moment_form_moment_form(
                                    moments, n1=n, n2=n_prime, l1=l, l2=l2, l3=l3, l4=l4, j=j
                                )
                                if verbose:
                                    print(f'{len(invariants)} {n} {l} {l2} ')
                                invariants.append(invariant)
    return invariants


def get_invariant_header(invariant):
    coeffs_dict = invariant.expand().evalf().as_coefficients_dict()
    parts = tuple(coeffs_dict)

    if isinstance(parts[0], sympy.Mul):
        terms = parts[0].args
    else:
        terms = [parts[0]]

    power = 0
    weight = 0
    for term in terms:
        print(term)
        if isinstance(term, sympy.Pow):
            exponent = term.args[1]
            power += exponent
            weight += exponent * term.args[0].indices[0]
        else:
            power += 1
            weight += term.indices[0]

    return f'{str(weight / 2)} {len(coeffs_dict)} {power}'


def get_string_block(invariant):
    coeffs_dict = invariant.expand().evalf().as_coefficients_dict()
    lines = [get_invariant_header(invariant)]

    for terms, coeff in coeffs_dict.items():
        parts = []
        parts.append(str(coeff))

        if isinstance(terms, sympy.Mul):
            pts = terms.args
        else:
            pts = [terms]

        for term in pts:
            if isinstance(term, sympy.Pow):
                symbol, power = term.args
                indices = symbol.indices
                parts.extend([f'{indices[0]} {indices[1]} {indices[2]}'] * power)
            else:
                indices = term.indices
                parts.append(f'{indices[0]} {indices[1]} {indices[2]}')
        lines.append(' '.join(parts))

    return '\n'.join(lines)



# -*- coding: utf-8 -*-
import numpy as np
import sympy

import milad
from milad import atomic
from milad import functions
from milad import invariants


def test_descriptor_derivatives():
    invs = invariants.read(invariants.COMPLEX_INVARIANTS, max_order=3)

    descriptor = milad.descriptor(
        features=dict(type=milad.functions.WeightedDelta, map_species_to=None),
        cutoff=5.,
        apply_cutoff=False,
        invs=invs
    )

    # Create a structure with just one atom, all symbolic
    r = sympy.IndexedBase('r', real=True)  # The positions
    s = sympy.IndexedBase('s', real=True)  # The species

    structure = atomic.AtomsCollection(
        1, positions=np.array([r[1], r[2], r[3]]), numbers=np.array([s[1]]), dtype=type(r[1])
    )
    phi, jac = descriptor(structure, jacobian=True)

    for i, entry in enumerate(phi):
        for j in range(3):
            diff = sympy.diff(entry, r[j + 1]).expand()
            from_jac = jac[i, j].expand()
            difference = diff - from_jac
            if not diff == from_jac:
                # If they differ, check that it's by a meaningful amount
                coeffs = np.array(tuple(difference.as_coefficients_dict().values()))
                np.testing.assert_array_almost_equal(coeffs, 0., decimal=10)



# -*- coding: utf-8 -*-
import abc
import functools
from typing import List, Callable, Union, Sequence

import numpy

import milad
from milad import geometric

__all__ = (
    'CutoffFunction', 'Distribution', 'Gaussian3D', 'GaussianEnvironment', 'SmoothGaussianEnvironment', 'cos_cutoff',
    'make_cutoff_params'
)

CutoffFunction = Callable[[float], float]


def cos_cutoff(cutoff: float, x: float):
    return 0.5 * (numpy.cos(numpy.pi * x / cutoff) + 1)


def make_cutoff_params(cutoff_fn: Callable[[float, float], float], cutoff: float):
    return {'cutoff': cutoff, 'cutoff_function': functools.partial(cutoff_fn, cutoff)}


def make_cutoff_function(name: str, cutoff: float = 6.):
    if name == 'cos':
        return functools.partial(cos_cutoff, cutoff)

    raise ValueError('Unknown cutoff function: {}'.format(name))


class Distribution(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def moment_tensor(self, max_order: int, normalise=False) -> numpy.array:
        """Calculate the moment tensor up to the given maximum order, optionally normalising.
        This will result in a numpy.array which has dimensions max_order * max_order * max_order"""

    def calc_moment_invariants(
        self, invariants: Sequence[milad.invariants.MomentInvariant], normalise=True
    ) -> numpy.array:
        """Given a sequence of invariants, calculate their values using the moments of this
        distribution"""
        # Figure out the maximum order of moment we will need for this set of invariants
        max_order = 0
        for inv in invariants:
            max_order = max(max_order, inv.max_order)

        raw_moments = self.moment_tensor(max_order, normalise)
        return numpy.fromiter((invariant.apply(raw_moments, normalise=normalise) for invariant in invariants),
                              dtype=numpy.float64)


class Gaussian3D(Distribution):
    """A simple 3D gaussian"""

    def __init__(self, pos: numpy.array, sigma: float, volume=1.):
        self.pos = pos
        self.sigma = sigma
        self.volume = volume

    def moment_tensor(self, max_order: int, normalise=False) -> numpy.array:
        return geometric.gaussian_geometric_moments(
            max_order=max_order, mu=self.pos, sigma=self.sigma, weight=self.volume
        )


class GaussianEnvironment(Distribution):
    """An environment made up of Gaussians"""

    def __init__(self):
        self._gaussians: List[Gaussian3D] = []

    def append(self, gaussian: Gaussian3D):
        self._gaussians.append(gaussian)

    def moment_tensor(self, max_order: int, normalise=False) -> numpy.array:
        """Calculate the raw moments of the collection of Gaussians"""
        moms = numpy.zeros((max_order + 1, max_order + 1, max_order + 1))
        for gaussian in self._gaussians:
            moms += gaussian.moment_tensor(max_order, normalise=normalise)
        return moms


class SmoothGaussianEnvironment(Distribution):
    """An environment with some spatial position and a cutoff sphere around it.  There can,
    optionally be a cutoff function applied"""

    def __init__(
        self,
        pos: numpy.array = numpy.zeros(3),
        cutoff: float = 6.,
        cutoff_function: Union[CutoffFunction, str] = None
    ):
        self._pos = pos
        self._cutoff_sq = cutoff * cutoff
        if isinstance(cutoff_function, str):
            self._cutoff_function = make_cutoff_function(cutoff_function)
        else:
            self._cutoff_function = cutoff_function
        self._gaussians = GaussianEnvironment()

    def add_gaussian(self, pos: numpy.array, sigma: float, weight=1.) -> Union[bool, float]:
        dr = pos - self._pos
        dist_sq = numpy.square(dr).sum()
        if dist_sq > self._cutoff_sq:
            return False

        if self._cutoff_function:
            weight *= self._cutoff_function(dist_sq**0.5)

        self._gaussians.append(Gaussian3D(dr, sigma, weight))
        return weight

    def add_gaussians(self, positions: numpy.array, sigma: float, mass=1.) -> Sequence[Union[bool, float]]:
        results = []
        for position in positions:
            results.append(self.add_gaussian(position, sigma, mass))

        return results

    def moment_tensor(self, max_order: int, normalise=False) -> numpy.array:
        return self._gaussians.moment_tensor(max_order, normalise)



# -*- coding: utf-8 -*-
import math
import numba
import numpy as np


def get_num_plane_repetitions_to_bound_sphere(radius: float, volume: float, cross_len: float) -> float:
    # The vector normal to the plane
    return radius / volume * cross_len


class AperiodicDistanceCalculator:

    def get_vecs_between(
        self,
        a: np.array,
        b: np.array,
        cutoff: float,
        max_cell_multiples: int = 100000,
        self_interation=True
    ) -> np.array:  # pylint: disable=invalid-name
        dr = b - a
        cutoff_sq = cutoff * cutoff

        out_values = []
        out_vec = dr
        if self_interation and np.dot(out_vec, out_vec) < cutoff_sq:
            out_values.append(out_vec)

        return np.array(out_values)


class UnitCellDistanceCalculator:

    def __init__(self, unit_cell: np.array):
        # Precompute some useful values
        self._cell = unit_cell
        self._a = self._cell[0]
        self._b = self._cell[1]
        self._c = self._cell[2]
        self._volume = np.abs(np.dot(self._a, np.cross(self._b, self._c)))

        self._a_cross_b = np.cross(self._a, self._b)
        self._a_cross_b_len = np.linalg.norm(self._a_cross_b)
        self._a_cross_b_hat = self._a_cross_b / self._a_cross_b_len

        self._b_cross_c = np.cross(self._b, self._c)
        self._b_cross_c_len = np.linalg.norm(self._b_cross_c)
        self._b_cross_c_hat = self._b_cross_c / self._b_cross_c_len

        self._a_cross_c = np.cross(self._a, self._c)
        self._a_cross_c_len = np.linalg.norm(self._a_cross_c)
        self._a_cross_c_hat = self._a_cross_c / self._a_cross_c_len

    def get_num_plane_repetitions_to_bound_sphere_planes(
        self, plane_vec1: np.array, plane_vec2: np.array, radius: float
    ) -> float:
        # The vector normal to the plane
        normal = np.cross(plane_vec1, plane_vec2)
        vol = self._volume  # = a . |b x c|
        return radius / vol * np.linalg.norm(normal)

    def get_vec_min_img(
        self,
        a: np.array,
        b: np.array,
        max_cell_multiples: int = 100000,
        self_interation=True
    ) -> np.array:  # pylint: disable=invalid-name
        """
        Get all vectors from a to b that are less than the cutoff in length

        :param a:
        :param b:
        :param cutoff:
        :param max_cell_multiples:
        :param self_interation:
        :return:
        """
        vol = self._volume
        # TODO: Wrap a, and b into the current unit cell
        dr = b - a

        min_dr_sq = np.dot(dr, dr)
        min_length = min_dr_sq**0.5

        a_max = math.ceil(get_num_plane_repetitions_to_bound_sphere(min_length, vol, self._b_cross_c_len))

        b_max = math.ceil(get_num_plane_repetitions_to_bound_sphere(min_length, vol, self._a_cross_c_len))

        c_max = math.ceil(get_num_plane_repetitions_to_bound_sphere(min_length, vol, self._a_cross_b_len))

        a_max = min(a_max, max_cell_multiples)
        b_max = min(b_max, max_cell_multiples)
        c_max = min(c_max, max_cell_multiples)

        # min_dr = fast_min_img(dr, self._a, self._b, self._c, a_max, b_max, c_max, self_interation)
        # return min_dr

        min_dr = dr

        for i in range(-a_max, a_max + 1):
            ra = i * self._a
            for j in range(-b_max, b_max + 1):
                rab = ra + j * self._b
                for k in range(-c_max, c_max + 1):
                    if not self_interation and i == 0 and j == 0 and k == 0:
                        continue

                    out_vec = rab + k * self._c + dr
                    len_sq = np.dot(out_vec, out_vec)
                    if len_sq < min_dr_sq:
                        min_dr = out_vec
                        min_dr_sq = len_sq

        return min_dr

    def get_vecs_between(
        self,
        a: np.array,
        b: np.array,
        cutoff: float,
        max_cell_multiples: int = 100000,
        self_interation=True
    ) -> np.array:  # pylint: disable=invalid-name
        """
        Get all vectors from a to b that are less than the cutoff in length

        :param a:
        :param b:
        :param cutoff:
        :param max_cell_multiples:
        :param self_interation:
        :return:
        """
        vol = self._volume
        # TODO: Wrap a, and b into the current unit cell
        dr = b - a

        a_max = math.floor(
            get_num_plane_repetitions_to_bound_sphere(
                cutoff + math.fabs(np.dot(dr, self._b_cross_c_hat)), vol, self._b_cross_c_len
            )
        )

        b_max = math.floor(
            get_num_plane_repetitions_to_bound_sphere(
                cutoff + math.fabs(np.dot(dr, self._a_cross_c_hat)), vol, self._a_cross_c_len
            )
        )

        c_max = math.floor(
            get_num_plane_repetitions_to_bound_sphere(
                cutoff + math.fabs(np.dot(dr, self._a_cross_b_hat)), vol, self._a_cross_b_len
            )
        )

        a_max = min(a_max, max_cell_multiples)
        b_max = min(b_max, max_cell_multiples)
        c_max = min(c_max, max_cell_multiples)

        cutoff_sq = cutoff * cutoff
        out_values = []

        for i in range(-a_max, a_max + 1):
            ra = i * self._a
            for j in range(-b_max, b_max + 1):
                rab = ra + j * self._b
                for k in range(-c_max, c_max + 1):
                    if not self_interation and i == 0 and j == 0 and k == 0:
                        continue

                    out_vec = rab + k * self._c + dr
                    if np.dot(out_vec, out_vec) < cutoff_sq:
                        out_values.append(out_vec)

        return np.array(out_values)


@numba.jit(parallel=True)
def fast_min_img(dr, a, b, c, a_max, b_max, c_max, self_interation):
    min_dr = dr.copy()
    min_dr_sq = np.dot(dr, dr)

    for i in range(-a_max, a_max + 1):
        ra = i * a
        for j in range(-b_max, b_max + 1):
            rab = ra + j * b
            for k in range(-c_max, c_max + 1):
                if not self_interation and i == 0 and j == 0 and k == 0:
                    continue

                out_vec = rab + k * c + dr
                len_sq = np.dot(out_vec, out_vec)
                if len_sq < min_dr_sq:
                    min_dr = out_vec
                    min_dr_sq = len_sq

    return min_dr



# -*- coding: utf-8 -*-
import random

import numpy as np
import sympy
from sympy import Symbol, IndexedBase

from milad import functions


def test_cosine_cutoff():
    # pylint: disable=invalid-name

    rcut = 1.5  # Larger than the random range so everything is within the cutoff sphere

    # Create some features
    delta = functions.WeightedDelta(np.random.rand(3), weight=random.random())
    gaussian = functions.WeightedGaussian(np.random.rand(3), sigma=random.random(), weight=random.random())
    features = functions.Features(delta, gaussian)

    cos_cut = functions.CosineCutoff(rcut)
    res, jac = cos_cut(features, jacobian=True)

    assert isinstance(res, functions.Features)
    # Order should be preserved
    res_delta = res.features[0]
    assert isinstance(res_delta, functions.WeightedDelta)
    assert np.all(res_delta.pos == delta.pos)
    assert res_delta.weight == delta.weight * cos_cut.fn(np.linalg.norm(delta.pos))

    res_gaussian = res.features[1]
    assert isinstance(res_gaussian, functions.WeightedGaussian)
    assert res_gaussian.weight == gaussian.weight * cos_cut.fn(np.linalg.norm(gaussian.pos))

    # Now let's check derivatives
    r = IndexedBase('r', real=True)
    r_c = IndexedBase('r_c', real=True)
    w = Symbol('w')
    dr = sympy.sqrt(r[0]**2 + r[1]**2 + r[2]**2)

    cut_expr = functions.CosineCutoff.symbolic(dr, w, r_c)
    # Create the numeric substitution for the delta function
    delta_subs = [(r[0], delta.pos[0]), (r[1], delta.pos[1]), (r[2], delta.pos[2]), (w, delta.weight), (r_c, rcut)]

    assert np.all(jac[0:3, 0:3] == np.eye(3))

    for i in range(3):
        np.testing.assert_almost_equal(jac[delta.WEIGHT, i], sympy.diff(cut_expr, r[i]).subs(delta_subs).evalf())
    np.testing.assert_approx_equal(jac[delta.WEIGHT, 3], sympy.diff(cut_expr, w).subs(delta_subs).evalf())

    gaussian_jac = jac[len(delta):, len(delta):]
    gaussian_subs = [(r[0], gaussian.pos[0]), (r[1], gaussian.pos[1]), (r[2], gaussian.pos[2]), (w, gaussian.weight),
                     (r_c, rcut)]

    assert np.all(gaussian_jac[0:3, 0:3] == np.eye(3))

    for i in range(3):
        np.testing.assert_almost_equal(
            gaussian_jac[gaussian.WEIGHT, i],
            sympy.diff(cut_expr, r[i]).subs(gaussian_subs).evalf()
        )

    np.testing.assert_approx_equal(
        gaussian_jac[gaussian.WEIGHT, gaussian.WEIGHT],
        sympy.diff(cut_expr, w).subs(gaussian_subs).evalf()
    )



# -*- coding: utf-8 -*-
import collections
from typing import Union, Sized, Iterator, Sequence, Optional, Tuple, List

import ase
import matplotlib.pyplot as plt
import numpy as np

__all__ = 'calculate_all_pair_distances', 'FingerprintSet'


def calculate_all_pair_distances(vectors, sort_result=True):
    """Calculate all pair distances between the given vectors"""
    num = len(vectors)
    lengths = []
    for i in range(num - 1):
        for j in range(i + 1, num):
            dr = vectors[i] - vectors[j]  # pylint: disable=invalid-name
            lengths.append(np.linalg.norm(dr))

    if sort_result:
        lengths.sort()
    return lengths


def even(val: int) -> bool:
    """Test if an integer is event.  Returns True if so."""
    return (val % 2) == 0


def inclusive(*args) -> Iterator[int]:
    """Like range() but inclusive of upper bound and automatically does iteration of ranges with a
    negative step e.g. 0, -4 will produce a range containing 0, -1, -2, -3, -4"""
    if len(args) not in (1, 2, 3):
        raise ValueError('Takes one or two args, got: {}'.format(args))

    if len(args) == 3:
        # Assume form is start, stop, step
        start, stop, step = args
    else:
        if len(args) == 1:
            start = 0
            stop = args[0]
        else:
            start = args[0]
            stop = args[1]

        step = 1 if start <= stop else -1

    sign = 1 if step > 0 else -1
    idx = start
    while sign * (stop - idx) >= 0:
        yield idx
        idx += step


class CoefficientCapture:

    class Capture:

        def __init__(self, mtx: np.array, idx):
            self._mtx = mtx
            self._idx = idx

        def __mul__(self, other):
            print('Coeff: {}, idx: {}'.format(other, self._idx))
            self._mtx[self._idx] = other
            return self

        def __rmul__(self, other):
            return self.__mul__(other)

        def __iadd__(self, other):
            self._mtx[self._idx] += other

        def __riadd__(self, other):
            return self.__iadd__(other)

    def __init__(self, shape: tuple):
        self._mtx = np.zeros(shape)

    def __getitem__(self, item):
        print('Capturing {}'.format(item))
        return self.Capture(self._mtx, item)

    @property
    def mtx(self) -> np.array:
        return self._mtx


def outer_product(*array) -> np.array:
    if not array:
        raise ValueError('No arrays supplied')

    product = array[0]
    for entry in array[1:]:
        product = np.tensordot(product, entry, axes=0)

    return product


SystemInfo = collections.namedtuple('SystemInfo', 'atoms fingerprints derivatives')


class FingerprintSet:
    """Container to store a set of fingerprints"""

    def __init__(self, fingerprint_length: int):
        self._fingerprint_length = fingerprint_length
        self._system_info: List[SystemInfo] = []

    def __len__(self):
        return len(self._system_info)

    @property
    def fingerprint_len(self) -> int:
        """Get the length of the fingerprint vector"""
        return self._fingerprint_length

    @property
    def total_environments(self) -> int:
        """Get the total number of environments summing up over all of the systems"""
        return sum(len(info.atoms) for info in self._system_info)

    @property
    def fingerprints(self) -> Sequence:
        """Access all the fingerprints"""
        return tuple(info.fingerprints for info in self._system_info)

    def has_all_derivatives(self):
        try:
            self.fingerprint_derivatives.index(None)
            return False
        except ValueError:
            # None not found
            return True

    @property
    def fingerprint_derivatives(self) -> Sequence:
        """Access all the fingerprint derivatives"""
        return tuple(info.derivatives for info in self._system_info)

    @property
    def sizes(self):
        return tuple(len(info.atoms) for info in self._system_info)

    def get_potential_energies(self, normalise=True) -> tuple:
        """Get the potential energies of all systems in the set"""
        if normalise:
            return tuple(info.atoms.get_potential_energy() / len(info.atoms) for info in self._system_info)

        return tuple(info.atoms.get_potential_energy() for info in self._system_info)

    def has_all_forces(self) -> bool:
        try:
            self.get_forces().index(None)
            return False
        except ValueError:
            # None not found
            return True

    def get_forces(self) -> Tuple[np.ndarray]:
        all_forces = []
        for info in self._system_info:
            try:
                forces = info.atoms.get_array('force')
            except KeyError:
                forces = info.atoms.get_forces()
            all_forces.append(forces)

        return tuple(all_forces)

    def add_system(self, system: ase.Atoms, fingerprints: Sized, derivatives: Optional[Sized] = None):
        """Add a system along with the fingerprints for all its environments"""
        if len(system) != len(fingerprints):
            raise ValueError(
                'There must be as many fingerprints as there are atoms, ' \
                'got {} atoms and {} environments'.format(len(system), len(fingerprints)))

        derivatives = np.array(derivatives) if derivatives is not None else None
        self._system_info.append(SystemInfo(system, fingerprints, derivatives))

    def systemwise_sum(self, values, normalise=True):
        """Given a vector of values, one per environment, this will sum up the values for each atomic system
        and return the result as a container whose length is the same as the number of systems

        The sum can optionally be normalised by the number of atoms in each system.
        """
        out = []
        idx = 0
        for info in self._system_info:
            natoms = len(info.atoms)
            summed = sum(values[idx:idx + natoms])
            if normalise:
                summed = summed / natoms
            out.append(summed)
            idx += natoms

        return out

    def plot_environments(self):
        """Create a plot of the environments.  Returns the matplotlib figure when can then be .show()n"""
        fig, axes = plt.subplots(figsize=(16, 5))

        for envs in self.fingerprints:
            for env in envs:
                axes.plot(tuple(range(len(env))), env, linewidth=1.0)

        return fig

    def plot_energies(self):
        fig, axes = plt.subplots(figsize=(16, 5))

        energies = self.get_potential_energies(normalise=True)
        axes.hist(energies, 50, log=True)
        axes.set_xlabel('Energy')
        axes.set_ylabel('Number')

        return fig

    def split(self, split_point: Union[float, int]) -> Tuple['FingerprintSet', 'FingerprintSet']:
        """Split this set into two.  This can be used for creating a training and validation set.

        The split point can be an integer, in which case it is treated as the index of the split point,
        or a float in which case is it treated as a ratio.
        """
        if isinstance(split_point, float):
            split_point = int(round(len(self) * split_point))
        elif not isinstance(split_point, int):
            raise TypeError('split_point must be integer or float, got {}'.format(split_point.__class__.__name__))

        # Split into two halves
        one = FingerprintSet(self.fingerprint_len)
        for info in self._system_info[:split_point]:
            one.add_system(*info)

        two = FingerprintSet(self.fingerprint_len)
        for info in self._system_info[split_point:]:
            two.add_system(*info)

        return one, two



# -*- coding: utf-8 -*-
from milad import generate
from milad import optimisers
from milad import zernike


def test_moments_optimiser(complex_invariants):
    """Basic tests for the moments optimiser"""
    pts = generate.random_points_in_sphere(10, 1.)

    # Create some Zernike moments from delta functions
    moms = zernike.from_deltas(complex_invariants.max_order, pts)
    # Create the invariants vector
    target = complex_invariants(moms)

    # Make settings very loose so this is fast, we're just testing the logic here...
    optimiser = optimisers.MomentsOptimiser()
    result = optimiser.optimise(
        invariants_fn=complex_invariants,
        target=target,
        initial=zernike.ZernikeMoments.rand(complex_invariants.max_order),
        target_rmsd=1e-2,
        cost_tol=1e-2,
        grad_tol=1e-2,
        verbose=False,
    )
    assert result.success, result.message



# -*- coding: utf-8 -*-
from . import analytic
from . import atomic
from . import invariants
from . import exceptions
from . import functions
from . import generate
from . import geometric
from . import models
from . import optimisers
from . import utils
from . import play
from . import plot
from . import reconstruct
from .play import asetools
from .atomic import *
from .fingerprinting import *
from .geometric import *
from .invariants import *
from .reconstruct import *
from .zernike import *

__all__ = ('analytic', 'atomic', 'invariants', 'functions', 'geometric', 'utils', 'play', 'plot', 'reconstruct',
           'zernike', 'exceptions', 'models', 'optimisers', 'generate', 'asetools') +\
          zernike.__all__ + fingerprinting.__all__ + reconstruct.__all__ + atomic.__all__ + invariants.__all__ + \
          geometric.__all__



# -*- coding: utf-8 -*-
import pytest

import numpy as np
import sympy

from milad import geometric
from milad import zernike
from milad import invariants
from milad import utils
from milad import generate
from milad import transform

# pylint: disable=invalid-name


def test_zernike_reconstruct_deltas():
    num_points = 4
    positions = generate.random_points_in_sphere(num_points, radius=.7)
    weights = 1.
    max_order = 7
    n_samples = 11

    moms = zernike.from_deltas(max_order, positions, weights=weights)
    reconstructed_values = moms.value_at(positions)

    # Now reconstruct a voxel grid
    spacing = np.linspace(-1., 1., n_samples)
    grid_points = []
    for pos in np.array(np.meshgrid(spacing, spacing, spacing)).reshape(3, -1).T:  # pylint: disable=not-an-iterable
        if np.linalg.norm(pos) > 1.:
            continue
        grid_points.append(pos)
    grid_points = np.array(grid_points)

    grid_vals = moms.value_at(grid_points)

    # The values at the original delta functions should hold the highest reconstruction values
    # (higher than anywhere else on the grid)
    assert grid_vals.max() < reconstructed_values.max()


def test_zernike_of_deltas(complex_invariants):
    positions = generate.random_points_in_sphere(4, radius=.7)
    weights = 1.
    max_order = 10

    moms = zernike.from_deltas(max_order, positions, weights=weights)
    invs_one = invariants.apply_invariants(complex_invariants, moms)

    # Now rotate the system randomly and recalculate the invariants
    rotated = transform.randomly_rotate(positions)
    rot_moms = zernike.from_deltas(max_order, rotated)
    invs_two = invariants.apply_invariants(complex_invariants, rot_moms)

    np.testing.assert_array_almost_equal(invs_two, invs_one)


def test_zernike_of_gaussians(complex_invariants):
    positions = generate.random_points_in_sphere(4, radius=.7)
    sigma = 0.1
    mass = 1.
    max_order = 10

    geom_moments = geometric.from_gaussians(max_order, positions, sigma, mass)
    zmoments = zernike.from_geometric_moments(max_order, geom_moments)
    invs_one = invariants.apply_invariants(complex_invariants, zmoments)

    # Now rotate the system randomly and recalculate the invariants
    rotated = transform.randomly_rotate(positions)
    rot_geom_moments = geometric.from_gaussians(max_order, rotated, sigma, mass)
    rot_zmoments = zernike.from_geometric_moments(max_order, rot_geom_moments)
    invs_two = invariants.apply_invariants(complex_invariants, rot_zmoments)

    np.testing.assert_array_almost_equal(invs_two, invs_one)


def test_zernike_properties():
    num_particles = 4
    positions = generate.random_points_in_sphere(num_particles, radius=.7)
    sigmas = np.random.rand(num_particles)
    weights = np.random.rand(num_particles)
    max_order = 10

    geom_moments = geometric.from_gaussians(max_order, positions, sigmas=sigmas, weights=weights)
    assert geom_moments[0, 0, 0] == weights.sum()

    moms = zernike.from_geometric_moments(max_order, geom_moments)
    assert moms[0, 0, 0] == 3 / (4 * np.pi) * weights.sum()

    for n in range(max_order + 1):
        for l in range(n + 1):
            if ((n - l) % 2) != 0:
                continue

            for m in range(l):
                assert moms[n, l, -m] == (-1)**m * moms[n, l, m].conjugate()


def test_zernike_analytic():
    max_order = 6
    geom = sympy.IndexedBase('m', real=True)  # Geometric moments
    jacobian = zernike.get_jacobian_wrt_geom_moments(max_order)

    # Let's build the Jacobian symbolically
    for _idx, (n, l, m) in enumerate(zernike.iter_indices(max_order, redundant=False)):
        eq = zernike.omega_nl_m(n, l, m, geom)
        for p, q, r in geometric.iter_indices(max_order):
            differentiated = eq.diff(geom[p, q, r])

            zindex = zernike.linear_index((n, l, m))
            gindex = geometric.linear_index(max_order, (p, q, r))
            jacobian_value = jacobian[zindex, gindex]

            assert complex(differentiated) == pytest.approx(jacobian_value), f'Omega{n},{l},{m} m{p},{q},{r}'


def test_zernike_indexing():
    max_order = 6

    # Check that linear indexing works correctly
    idx = 0
    for n in utils.inclusive(max_order):
        for l in utils.inclusive(n):
            if not utils.even(n - l):
                continue

            for m in utils.inclusive(l):
                assert zernike.linear_index((n, l, m)) == idx
                idx += 1

            for m in utils.inclusive(-l, -1, 1):
                assert zernike.linear_index((n, l, m)) == idx
                idx += 1


def test_zernike_from_vector():
    """Test converting Zernike moments to a vector and back"""
    num_points = 4
    positions = generate.random_points_in_sphere(num_points, radius=.7)
    max_order = 7

    moms = zernike.from_deltas(max_order, positions)
    as_vector = moms.vector

    from_vector = zernike.ZernikeMoments.from_vector(max_order, as_vector)
    assert np.allclose(from_vector.vector, as_vector)

    # Double check the round-trip
    assert np.all(zernike.ZernikeMoments.from_vector(max_order, from_vector.vector).vector == from_vector.vector)


def test_zernike_builder():
    """Test that the Zernike builder can create the moments from a vector and the inverse"""
    num_points = 4
    positions = generate.random_points_in_sphere(num_points, radius=.7)
    max_order = 7

    moms = zernike.from_deltas(max_order, positions)
    builder = moms.get_builder()

    # Get the moments in vector form
    vec = builder.inverse(moms)

    # Now let's create a set of moments from the vector and check that it's all the same
    recreated, _jac = builder(vec, jacobian=True)
    assert recreated == moms



# -*- coding: utf-8 -*-
import matplotlib.pyplot
import numpy


def plot_invariants(values,
                    axes: matplotlib.pyplot.Axes = None,
                    label=None,
                    scatter=True):
    if axes is None:
        _fig, axes = matplotlib.pyplot.subplots()

    kwargs = {}
    if label is not None:
        kwargs['label'] = label

    axes.plot(numpy.array(range(len(values))), values, **kwargs)

    if scatter:
        axes.scatter(numpy.array(range(len(values))), values, s=5.)

    return axes


def plot_multiple_invariants(series,
                             axes: matplotlib.pyplot.Axes = None,
                             labels='auto'):
    if axes is None:
        _fig, axes = matplotlib.pyplot.subplots()

    if labels == 'auto':
        labels = tuple('Series {}'.format(idx) for idx in range(len(series)))

    for idx, series in enumerate(series):
        kwargs = {}
        if labels is not None:
            kwargs['label'] = labels[idx]

        axes.plot(numpy.array(range(len(series))), series, **kwargs)

    axes.set_xlabel('Moment invariant #')
    axes.set_ylabel('Invariant value')
    return axes



# -*- coding: utf-8 -*-
import collections
import math
from typing import Tuple, Optional, Callable

import numpy as np
from scipy import optimize

from milad import functions
from . import utils

__all__ = ('LeastSquaresOptimiser',)

OptimiserResult = collections.namedtuple('OptimiserResult', 'success message value rmsd n_func_eval n_jac_eval')

BoundsType = Tuple[Optional[functions.StateLike], Optional[functions.StateLike]]


class LeastSquaresOptimiser:
    """Uses scipy optimize.least_squares to perform optimisation"""

    def optimise(  # pylint: disable=too-many-locals
            self,
            func: functions.Function,
            initial: functions.StateLike,
            mask: functions.StateLike = None,
            jacobian='native',
            bounds: BoundsType = (-np.inf, np.inf),
            max_func_evals=10000,
            x_tol=1e-8,
            cost_tol=1e-6,
            grad_tol=1e-8,
            callback: Callable = None,
            verbose=False,
    ) -> OptimiserResult:
        """
        :param func: the function to optimise
        :param initial: the initial state
        :param mask: a mask of pre-set values whose degrees of freedom will not be present in the optimisation vector
        :param jacobian: if 'native' the analytic Jacobian will be requested from 'func', otherwise this option is
            passed to optimize.least_squares
        :param bounds: place bounds on the possible inputs to func
        :param max_func_evals: the maximum number of function evaluations.  If not specified will be 50 * len(initial).
        :param x_tol: tolerance for termination by change of the independent variables.
        :param cost_tol: tolerance in change of cost function.  If the difference between one step and another goes
            below this value the optimisation will stop
        :param verbose: print during the optimisation

        :return:
        """
        # SciPy can only deal with numpy arrays so if we get one, all good, otherwise we have to rely on a builder
        if isinstance(initial, np.ndarray):
            fun = func
            x0_ = initial
            builder = None
        else:
            # Need to add a builder step
            builder = initial.get_builder(mask=mask)
            fun = functions.Chain(builder, func)
            x0_ = builder.inverse(initial)
            bounds = list(bounds)
            # Convert the bounds
            if isinstance(bounds[0], type(initial)):
                bounds[0] = builder.inverse(bounds[0])
            if isinstance(bounds[1], type(initial)):
                bounds[1] = builder.inverse(bounds[1])
            bounds = tuple(bounds)

        complex_input = False
        if np.iscomplexobj(x0_):
            complex_input = True
            x0_ = utils.split_state(x0_)

        # Annoyingly scipy least_squares calls the function and the separately
        # but we don't support getting the Jacobian on it's own (it always comes
        # with a function evaluation and we don't want to call it twice so just
        # cache the result)
        jac = self._jac if jacobian == 'native' else jacobian
        data = utils.Data(
            use_jacobian=jacobian == 'native',
            complex_input=complex_input,
            verbose=verbose,
            builder=builder,
            callback=callback,
        )

        max_func_evals = max_func_evals if max_func_evals is not None else 100 * len(initial)

        if np.isnan(x0_).any():
            raise ValueError()

        # Do it!
        res = optimize.least_squares(
            self._calc,
            x0_,
            jac=jac,
            kwargs=dict(func=fun, opt_data=data),
            bounds=bounds,
            xtol=x_tol,
            ftol=cost_tol,
            gtol=grad_tol,
            max_nfev=max_func_evals,
        )

        if complex_input:
            value = utils.join_state(res.x)
        else:
            value = res.x

        if builder is not None:
            value = builder(value)

        # Now convert the result to our optimiser result format
        return OptimiserResult(
            success=res.success,
            value=value,
            rmsd=math.sqrt(2. / len(res.x) * res.cost),
            message=res.message,
            n_func_eval=res.nfev,
            n_jac_eval=res.njev,
        )

    def optimise_target(
        self,
        func: functions.Function,
        initial: functions.State,
        target: functions.StateLike,
        mask: functions.State = None,
        jacobian='native',
        bounds: BoundsType = (-np.inf, np.inf),
        max_func_evals=10000,
        x_tol=1e-8,
        cost_tol=1e-6,
        grad_tol=1e-8,
        callback: Callable = None,
        verbose=False,
    ) -> OptimiserResult:
        """
        Optimise the function to a given target value.  This minimises the loss of the difference between the function
        and the target.

        :param func: the function to optimise
        :param initial: the initial state
        :param target: the output state
        :param jacobian: if 'native' the analytic Jacobian will be requested from 'func', otherwise this option is
            passed to optimize.least_squares
        :param bounds: place bounds on the possible inputs to func
        :param max_func_evals: the maximum number of force evaluations.  If not specified will be 50 * len(initial).
        :param x_tol: tolerance for termination by change of the independent variables.
        :param cost_tol: tolerance in change of cost function.  If the difference between one step and another goes
            below this value the optimisation will stop
        :param verbose: print during the optimisation
        :return:
        """
        # Calculate residuals to a particular target
        return self.optimise(
            functions.Chain(func, functions.Residuals(target)),
            initial=initial,
            mask=mask,
            jacobian=jacobian,
            bounds=bounds,
            x_tol=x_tol,
            cost_tol=cost_tol,
            grad_tol=grad_tol,
            max_func_evals=max_func_evals,
            callback=callback,
            verbose=verbose,
        )

    @staticmethod
    def _calc(state: functions.StateLike, func: functions.Function, opt_data: 'utils.Data') -> np.ndarray:
        if np.isnan(state).any():
            raise ValueError(f'iter: {opt_data.iter}, state: {state}')

        if opt_data.complex_input:
            state = utils.join_state(state)

        value = functions.get_bare(func(state))
        if opt_data.verbose:
            print('|Max| {}'.format(np.abs(value).max()))

        if opt_data.callback is not None:
            if opt_data.builder is None:
                inp = state
            else:
                inp = opt_data.builder(state)
            opt_data.callback(inp, value, None)

        split = utils.split_state(value)

        if np.isnan(split).any():
            raise ValueError(f'iter: {opt_data.iter}, split: {split}, value: {value}, state: {state}')

        opt_data.iter += 1

        return split

    @staticmethod
    def _jac(state: functions.StateLike, func: functions.Function, opt_data: 'utils.Data') -> np.ndarray:
        if np.isnan(state).any():
            raise ValueError()

        if opt_data.complex_input:
            state = utils.join_state(state)

        value, jac = func(state, jacobian=True)
        if opt_data.verbose:
            print('|Max| {}'.format(np.abs(value).max()))

        if opt_data.callback is not None:
            if opt_data.builder is None:
                inp = state
            else:
                inp = opt_data.builder(state)
            opt_data.callback(inp, value, jac)

        split = utils.split_jacobian(jac, complex_inputs=opt_data.complex_input, complex_outputs=np.iscomplexobj(value))

        if np.isnan(split).any():
            raise ValueError()

        return split



# -*- coding: utf-8 -*-
import numpy as np

import ase.build
import ase.visualize

import milad
from milad import functions
from milad import zernike
from milad.play import asetools


def test_structure_optimiser_invariants(complex_invariants):
    """Basic tests of the structure optimiser that is given a set of invariants"""
    radius_factor = 1.25
    molecule = ase.build.molecule('C3H9N')
    max_radius = asetools.prepare_molecule(molecule)
    milad_molecule = asetools.ase2milad(molecule)
    cutoff = radius_factor * max_radius

    descriptor = milad.descriptor(
        species={'map': {
            'numbers': tuple(molecule.numbers),
            'range': (1., 5.)
        }},
        features={
            'type': functions.WeightedDelta,
            'map_species_to': 'WEIGHT'
        },
        cutoff=cutoff,
        invs=complex_invariants,
        moments_calculator=zernike.ZernikeMomentCalculator(complex_invariants.max_order),
        apply_cutoff=False,
    )
    fingerprint = descriptor(milad_molecule)

    optimiser = milad.optimisers.StructureOptimiser()
    initial = milad_molecule.copy()
    initial.positions += 0.1 * (np.random.rand(*initial.positions.shape) - 0.5)
    result = optimiser.optimise(
        descriptor=descriptor,
        target=fingerprint,
        initial=initial,
        x_tol=1e-5,
        verbose=False,
    )
    assert result.rmsd < 5e-3


def test_structure_optimiser_moments(complex_invariants):
    """Basic tests of the structure optimiser that is given a set of moments"""
    radius_factor = 1.25
    molecule = ase.build.molecule('C3H9N')
    max_radius = asetools.prepare_molecule(molecule)
    milad_molecule = asetools.ase2milad(molecule)
    cutoff = radius_factor * max_radius

    descriptor = milad.descriptor(
        species={'map': {
            'numbers': tuple(molecule.numbers),
            'range': (0.5, 5.)
        }},
        features={
            'type': functions.WeightedDelta,
            'map_species_to': 'WEIGHT'
        },
        cutoff=cutoff,
        invs=complex_invariants,
        moments_calculator=zernike.ZernikeMomentCalculator(complex_invariants.max_order),
        apply_cutoff=False,
    )
    target = descriptor.get_moments(milad_molecule)

    optimiser = milad.optimisers.StructureOptimiser()
    initial = milad_molecule.copy()
    initial.positions += 0.1 * (np.random.rand(*initial.positions.shape) - 0.5)

    # Let's test the mask by removing the species degrees of freedom
    mask = initial.get_mask()
    mask.numbers = initial.numbers

    result = optimiser.optimise(
        descriptor=descriptor,
        target=target,
        initial=initial,
        x_tol=1e-4,
        mask=mask,
        verbose=False,
    )
    assert result.rmsd < 1e-3



# -*- coding: utf-8 -*-
import argparse
import functools
import collections
import logging
import random
from typing import List

import numpy as np
import rmsd as rmsdlib
from scipy.spatial import distance
from scipy import optimize
from sklearn import cluster

from . import atomic
from . import base_moments
from . import exceptions
from . import fingerprinting
from . import optimisers

__all__ = ('Decoder',)

_LOGGER = logging.getLogger(__name__)

StructureOptimisationResult = collections.namedtuple('StructureOptimisationResult', 'success atoms message rmsd')


@functools.singledispatch
def find_clusters(spec, num_clusters: int, **kwargs) -> np.ndarray:
    """Find clusters based on moments"""
    raise TypeError(f'Cannot find clusters from {spec.__class__.__name__}')


@find_clusters.register(base_moments.Moments)
def _(moments: base_moments.Moments, num_clusters: int, query: base_moments.ReconstructionQuery,
      fingerprinter: fingerprinting.MomentInvariantsDescriptor) -> np.ndarray:
    """Find clusters from moments.  This will take the moments and reconstruct values on a grid
    which will be used for the actual cluster determination"""
    # Calculate the grid values
    values = moments.reconstruct(query, zero_outside_domain=True)
    # Find the cluster centres
    return find_clusters((query.points, values), num_clusters)


@find_clusters.register(tuple)
def _(grid, num_clusters: int) -> np.ndarray:
    """Given a grid consisting of points and weights this function funds clusters using k-means"""
    grid_points, grid_values = grid
    kmeans = cluster.KMeans(num_clusters, max_iter=1000, algorithm='full', tol=1e-1)
    kmeans.fit(grid_points, sample_weight=grid_values)

    num_labels = len(set(kmeans.labels_))
    if num_labels != num_clusters:
        raise exceptions.ReconstructionError(f'Could not find {num_clusters} clusters, found {num_labels} labels')
    if len(kmeans.cluster_centers_) != num_clusters:
        raise exceptions.ReconstructionError(
            f'Could not find {num_clusters} clusters, found {len(kmeans.cluster_centers_)} centres'
        )

    return kmeans.cluster_centers_


def find_maximum(moments: base_moments.Moments, initial: np.ndarray) -> np.ndarray:
    """Find the maximum of a set of moments"""

    def calc(pos: np.ndarray):
        return -moments.value_at(pos)

    res = optimize.minimize(calc, initial)
    return res.x


def find_peaks(
    num_peaks: int,
    moments: base_moments.Moments,
    descriptor: fingerprinting.MomentInvariantsDescriptor,
    query: base_moments.ReconstructionQuery = None,
    grid_size=31,
):
    query = query or moments.create_reconstruction_query(moments.get_grid(grid_size), moments.max_order)
    current_grid = moments.reconstruct(query, zero_outside_domain=True)
    return find_peaks_from_grid(num_peaks, current_grid, descriptor, query)


class FoundPeaks:

    def __init__(self, initial: base_moments.Moments):
        self._initial = initial
        self._moments = []
        self._atom_positions = []

    @property
    def positions(self) -> List[np.ndarray]:
        return self._atom_positions

    @property
    def moments(self) -> List[base_moments.Moments]:
        return self._moments

    def append(self, pos: np.ndarray, moments: base_moments.Moments):
        self._atom_positions.append(pos)
        self._moments.append(moments)

    def remaining_value(self, pos: np.ndarray) -> float:
        initial = self._initial.value_at(pos)
        for atom_moments in self._moments:
            initial -= atom_moments.value_at(pos)
        return initial

    def find_next_maximum(self, start_pt: np.ndarray):
        res = optimize.minimize(self.remaining_value, start_pt)
        return res.x


def find_peaks_from_maxima(
    num_peaks: int,
    moments: base_moments.Moments,
    descriptor: fingerprinting.MomentInvariantsDescriptor,
    query: base_moments.ReconstructionQuery = None,
    grid_size=31,
):
    # pylint: disable=too-many-locals
    scale = 1.0
    if descriptor.scaler is not None:
        scale = 1. / descriptor.cutoff

    query = query or moments.create_reconstruction_query(moments.get_grid(grid_size), moments.max_order)
    peaks = FoundPeaks(moments)
    current_grid = moments.reconstruct(query, zero_outside_domain=True)
    outside = get_buffer_indices(query, current_grid)
    current_grid[outside] = 0.

    for _ in range(num_peaks):
        # Find the index of the maximum value in the current grid
        max_idx = current_grid.argmax()

        # Get that position in the grid
        guess = query.points[max_idx]
        # Now find the local maximum
        moments_pos = peaks.find_next_maximum(guess)

        # Build an atoms collection with a single atom at that position
        single_atom = atomic.AtomsCollection(1, positions=[moments_pos / scale], numbers=[1.])

        # Get the moments so we can subtract this from the grid
        single_moments = descriptor.get_moments(single_atom, preprocess=False)
        peaks.append(moments_pos, single_moments)

        # Subtract off the grid
        # Get the grid for just that atom on its own
        atom_grid = single_moments.reconstruct(query, zero_outside_domain=True)
        # Subtract off the single atom grid
        current_grid -= atom_grid

        current_grid[outside] = 0.

    return peaks.positions / scale


def find_peaks_from_grid(
    num_peaks: int,
    grid,
    descriptor: fingerprinting.MomentInvariantsDescriptor,
    query: base_moments.ReconstructionQuery,
    exclude_radius=0.51,
):
    # pylint: disable=too-many-locals
    subtract_signal = True
    scale = 1.0 if descriptor.cutoff is None else 1. / descriptor.cutoff

    current_grid = grid.copy()

    exclude = exclude_radius * scale
    outside = get_buffer_indices(query, current_grid)

    mask = np.zeros(current_grid.shape, dtype=bool)
    mask[outside] = True

    found_positions = []
    for _ in range(num_peaks):
        # Find the index of the maximum value in the current grid
        max_idx = current_grid.argmax()

        # Get that position in the grid
        atom_pos = query.points[max_idx]
        found_positions.append(atom_pos / scale)

        if subtract_signal:
            # Build an atoms collection with a single atom at that position
            single_atom = atomic.AtomsCollection(1, positions=[atom_pos / scale], numbers=[1.])

            # Get the moments so we can subtract this from the grid
            single_moments = descriptor.get_moments(single_atom, preprocess=False)

            # Get the grid for just that atom on its own
            atom_grid = single_moments.reconstruct(query, zero_outside_domain=True)

            # Subtract off the single atom grid
            current_grid -= atom_grid

        # Now mask off the gridpoints associated with this atom
        mask |= get_surrounding_gridpoints(query, atom_pos, exclude)

        current_grid[mask] = 0.

    return np.array(found_positions)


def get_surrounding_gridpoints(
    query: base_moments.ReconstructionQuery, atom_pos: np.array, radius: float
) -> np.ndarray:
    # First find the points within a cube of the central point
    mask = (query.points[:, 0] > atom_pos[0] - radius) & \
           (query.points[:, 0] < atom_pos[0] + radius) & \
           (query.points[:, 1] > atom_pos[1] - radius) & \
           (query.points[:, 1] < atom_pos[1] + radius) & \
           (query.points[:, 2] > atom_pos[2] - radius) & \
           (query.points[:, 2] < atom_pos[2] + radius)

    indices = np.argwhere(mask)[:, 0]
    rsq = radius * radius

    # Now check the gridpoints left to see if they fall within the cutoff sphere
    for idx in indices:
        dr = query.points[idx] - atom_pos  # pylint: disable=invalid-name
        if np.dot(dr, dr) > rsq:
            # Remove it from the mask
            mask[idx] = False

    return mask


def get_buffer_indices(query: base_moments.ReconstructionQuery, grid_values) -> List[int]:
    # Get all grid values that are currently 0, these should stay zero
    outside = list(np.argwhere(grid_values == 0.).reshape(-1))

    # Now, get those that are near the bounds of the sphere i.e. 1.0
    cutoff_sq = 0.85**2
    for idx, pt in enumerate(query.points):  # pylint: disable=invalid-name
        if np.sum(pt**2) > cutoff_sq:
            outside.append(idx)

    return outside


def find_atoms(
    # pylint: disable=too-many-locals
    num_atoms: int,
    moments: base_moments.Moments,
    descriptor: fingerprinting.MomentInvariantsDescriptor,
    query: base_moments.ReconstructionQuery = None,
    grid_size=31,
) -> atomic.AtomsCollection:
    """
    Given a set of moments this will try to extract atom positions and species by successively placing a new atom
    on the current peak in reconstruction grid.  After this the a new grid is calculated and subtracted from the
    original.

    :param moments:
    :param descriptor:
    :param num_atoms: the number of atoms to find
    :param query: the reconstruction query to use.
    :param grid_size: the size of reconstruction grid to use.  Only used if no reconstruction query is passed.
    :return:
    """
    query = query or moments.create_reconstruction_query(moments.get_grid(grid_size), moments.max_order)

    atom_positions = []
    atom_numbers = []
    optimiser = optimisers.StructureOptimiser()

    # Calculate the original moments grid
    orig_grid = moments.reconstruct(query, zero_outside_domain=True)
    current_grid = orig_grid

    preprocess = False

    while True:
        # Find the index of the maximum value in the current grid
        max_idx = current_grid.argmax()

        # Get that position in the grid
        atom_positions.append(query.points[max_idx] / descriptor.cutoff)
        atom_numbers.append(1.)

        # Build an atoms collection with the current set of atoms
        current_atoms = atomic.AtomsCollection(len(atom_positions), positions=atom_positions, numbers=atom_numbers)

        # Locally optimise the atomic positions
        res = optimiser.optimise(
            descriptor,
            target=moments,
            initial=current_atoms,
            preprocess=preprocess,
        )

        if num_atoms == current_atoms.num_atoms:
            # Reached the number of atoms limit
            break

        # Get the moments so we can subtract this from the grid
        current_moments = descriptor.get_moments(res.value, preprocess=preprocess)

        # Get the grid for this set of atoms
        this_grid = current_moments.reconstruct(query, zero_outside_domain=True)

        # Subtract off the current grid
        current_grid = orig_grid - this_grid

        # Now remove the signal of the atom from the grid
        remove_idxs = np.argwhere(current_grid >= (0.5 * current_grid.max()))
        current_grid[remove_idxs] = 0.

        atom_positions = list(res.value.positions)
        atom_numbers = list(res.value.numbers)

    return res.value


def create_atoms_collection(clusters: cluster.KMeans, atomic_numbers=1.):
    """Take a set of clusters and use the centres to construct an atoms collection"""
    num_atoms = len(clusters.cluster_centers_)
    return atomic.AtomsCollection(num_atoms, positions=clusters.cluster_centers_, numbers=atomic_numbers)


DecoderResult = collections.namedtuple(
    'DecoderResult', 'success message value rmsd moments_reconstruction initial_reconstruction atoms_reconstruction'
)


class Decoder:
    """This class decodes a structure from a set of moment invariants"""

    def __init__(
        self,
        descriptor: fingerprinting.MomentInvariantsDescriptor,
        query: base_moments.ReconstructionQuery,
    ):
        self._descriptor = descriptor
        self._query = query

        self._optimiser = optimisers.StructureOptimiser()
        self._moments_optimiser = optimisers.MomentsOptimiser()
        self._structure_optimiser = optimisers.StructureOptimiser()

    def decode(self, fingerprint: np.ndarray, num_atoms: int, verbose=False, get_trajectory=False) -> DecoderResult:
        decode_result = argparse.Namespace()

        # Generate an initial configuration using a structure with randomly placed atoms
        initial_atoms = self.create_random_atoms(num_atoms)

        # Reconstruct the moments
        result = self._moments_optimiser.optimise(
            self._descriptor.invariants, target=fingerprint, initial=self._descriptor.get_moments(initial_atoms)
        )

        decode_result.moments_reconstruction = result

        if verbose:
            print(f'm: {result.rmsd}, ', end='')

        initial_atoms = self.create_initial_atoms(result.value, num_atoms)

        # Now let's get optimise the atomic configuration to be consistent with the reconstructed moments
        result = self._structure_optimiser.optimise(
            self._descriptor,
            initial=initial_atoms,
            target=result.value,  # Target the found moments
            get_trajectory=get_trajectory,
        )

        decode_result.initial_reconstruction = result

        if verbose:
            print(f'sm: {result.rmsd}, ', end='')

        # Finally, optimise the found structure wrt to the fingerprint
        result = self._structure_optimiser.optimise(
            self._descriptor,
            initial=result.value,
            target=fingerprint,  # Target the fingerprint
            get_trajectory=get_trajectory,
        )

        if verbose:
            print(f'sf: {result.rmsd}, ', end='')

        decode_result.atoms_reconstruction = result
        decode_result.success = result.success
        decode_result.message = result.message
        decode_result.value = result.value
        decode_result.rmsd = result.rmsd

        return DecoderResult(**decode_result.__dict__)

    def create_random_atoms(self, num: int) -> atomic.AtomsCollection:
        """Create a random atoms configuration with the number of atoms passed"""
        return atomic.random_atom_collection_in_sphere(
            num,
            radius=self._descriptor.cutoff,
            numbers=random.choices(self._descriptor.species, k=num),
            centre=True,
        )

    def create_initial_atoms(self, moments: base_moments.Moments, num: int) -> atomic.AtomsCollection:
        # atoms = find_atoms(num, moments, self._descriptor, self._query)
        # atoms.numbers[:] = random.choices(self._descriptor.species, k=num)
        # return atoms

        peaks = find_peaks(num, moments, self._descriptor, self._query)
        return atomic.AtomsCollection(num, peaks, random.choices(self._descriptor.species, k=num))


def merge_atoms(system: atomic.AtomsCollection, dist_threshold=0.2):
    # Proceed to merging of atoms
    dists = distance.cdist(system.positions, system.positions)
    np.fill_diagonal(dists, np.inf)  # Get rid of diagonals as this is just the self-interaction
    merge_sets = []
    for (i, j) in np.argwhere(dists < dist_threshold):
        if i > j:
            # Ignore lower diagonal
            continue

        merged = False
        for merge_set in merge_sets:
            if i in merge_set or j in merge_set:
                if i in merge_set:
                    merge_set.add(j)
                elif j in merge_set:
                    merge_set.add(i)
                merged = True
                break

        if not merged:
            # Create a new merge set
            merge_sets.append({i, j})

    # Start merging
    positions = []
    numbers = []
    merged_indices = set()
    for merge_set in merge_sets:
        pos = np.zeros(3)
        number = 0.
        for i in merge_set:
            pos += system.positions[i]
            number += system.numbers[i]
            merged_indices.add(i)
        pos /= len(merge_set)

        positions.append(pos)
        numbers.append(number)

    for i in set(range(len(system.numbers))) - merged_indices:
        positions.append(system.positions[i])
        numbers.append(system.numbers[i])

    return len(merged_indices), atomic.AtomsCollection(len(positions), positions, numbers)


def get_best_rms(
    reference: atomic.AtomsCollection,
    probe: atomic.AtomsCollection,
    max_attempts: int = 1000,
    max_retries=20,
    threshold=1e-7,
) -> float:
    """
    Get the best RMSs fitting between two molecules.  This will first use an algorithm to make a decent guess at the
    best permutational ordering of atoms and then try a brute force search.

    :param reference: the reference to fit to
    :param probe: the probe to try to fit to the reference
    :return: the best RMSD found
    """
    from rdkit.Chem import rdMolAlign
    from . import rdkittools

    # Find a decent re-ordering of the atoms
    try:
        reorder_map = rmsdlib.reorder_hungarian(reference.numbers, probe.numbers, reference.positions, probe.positions)
    except ValueError:
        pass
    else:
        probe = atomic.AtomsCollection(
            reference.num_atoms,
            positions=probe.positions[reorder_map],
            numbers=probe.numbers[reorder_map],
        )

    # Now ask rdkit to find the best.
    # This will try a brute-force permutation search starting with the one we just defined which is likely to be
    # the best
    reference = rdkittools.milad2rdkit(reference)
    probe = rdkittools.milad2rdkit(probe)
    try:
        best = np.inf
        for _ in range(max_retries):
            best = min(rdMolAlign.GetBestRMS(probe, reference, maxMatches=max_attempts), best)
            if best < threshold:
                break
        return best
    except RuntimeError:
        return np.inf


def find_iteratively(
    descriptor: fingerprinting.MomentInvariantsDescriptor,
    fingerprint: np.ndarray,
    num_atoms: int,
    initial: atomic.AtomsCollection,
    find_species=False,
    verbose=False,
    min_rmsd=1e-7,
    max_iters=6,
    grid_query=None
):
    # pylint: disable=too-many-locals
    # Initialisation
    moments_optimiser = optimisers.MomentsOptimiser()
    structure_optimiser = optimisers.StructureOptimiser()
    atoms = initial

    mask = None
    if not find_species:
        # Fix the species numbers
        mask = atoms.get_mask()
        mask.numbers = atoms.numbers

    for i in range(max_iters):
        # Now recreate the moments from the atoms
        moments = descriptor.get_moments(atoms)

        if grid_query is None:
            # Create the reconstruction query the first time
            grid_query = moments.create_reconstruction_query(moments.get_grid(31), moments.max_order)

        result = moments_optimiser.optimise(
            invariants_fn=descriptor.invariants,
            target=fingerprint,
            initial=moments,
            verbose=False,
            # cost_tol=1e-5,
        )
        moments = result.value

        if verbose:
            print(f'{i} moms->fingerprint: {result.rmsd}')

        # Find the peaks and create the corresponding collection of atoms
        peaks = find_peaks(num_atoms, moments, descriptor, grid_query)
        atoms = atomic.AtomsCollection(num_atoms, peaks, numbers=initial.numbers)

        result = structure_optimiser.optimise(
            descriptor,
            target=moments,
            initial=atoms,
            mask=mask,
            verbose=False,
        )

        if verbose:
            print(f'{i} atoms->moms: {result.rmsd}')

        result = structure_optimiser.optimise(
            descriptor,
            target=fingerprint,
            initial=result.value,
            mask=mask,
            verbose=False,
        )

        # result = structure_optimiser.optimise(
        #     descriptor,
        #     target=fingerprint,
        #     initial=atoms,
        #     mask=mask,
        #     verbose=False,
        # )

        if verbose:
            print(f'{i} atoms->fingerprint: {result.rmsd}')

        if find_species:
            # Take the current result, fix the species and allow positions to vary
            atoms = result.value

            pos_mask = atoms.get_mask()
            pos_mask.numbers = atoms.numbers

            result = structure_optimiser.optimise(
                descriptor,
                target=fingerprint,
                initial=atoms,
                mask=pos_mask,
                verbose=False,
            )

            if verbose:
                print(f'{i} atom pos->fingerprint: {result.rmsd}')

        if result.rmsd < min_rmsd:
            break

        atoms = result.value

    return result



# -*- coding: utf-8 -*-
import argparse
import collections
import functools
from typing import Tuple, Union, List

import numpy as np

from milad import atomic
from milad import base_moments
from milad import fingerprinting
from milad import functions
from milad.play import asetools
from . import least_squares

__all__ = ('StructureOptimiser', 'StructureOptimisationResult')

StructureOptimisationResult = collections.namedtuple(
    'StructureOptimisationResult', 'success message value rmsd n_func_eval n_jac_eval traj'
)


class StructureOptimiser:
    """
    The structure optimiser takes a fingerprinting function and performs a least squares optimisation
    to match a structure to a given fingerprint.
    """

    def __init__(self):
        self._least_squares_optimiser = least_squares.LeastSquaresOptimiser()

    def optimise(
        # pylint: disable=too-many-locals
        self,
        descriptor: fingerprinting.MomentInvariantsDescriptor,
        target: Union[np.ndarray, base_moments.Moments],
        initial: atomic.AtomsCollection,
        jacobian='native',
        mask: atomic.AtomsCollection = None,
        bounds: Tuple[atomic.AtomsCollection, atomic.AtomsCollection] = None,
        x_tol=1e-8,
        cost_tol=1e-5,
        grad_tol=1e-8,
        max_func_evals=5000,
        preprocess=True,
        preprocessor: functions.Function = None,
        get_trajectory=False,
        verbose=False,
    ) -> StructureOptimisationResult:
        """
        :param descriptor: the descriptor with the settings used to generate the fingerprint
        :param target: the fingerprint to decode back into an atoms collection
        :param initial: the starting atoms configuration
        :param x_tol: tolerance for termination by the change of independent variables.
        :param cost_tol: stopping criterion for the fitting algorithm
        :param max_func_evals: the maximum number of allowed fingerprint evaluations
        :param preprocess: if True will apply the descriptor preprocessor to initial and the mask and then
            'un-preprocess' the output
        :param preprocessor: the function used to perform preprocessing.  The following inputs (if present) will be
            preprocessed:
                * initial
                * mask
                * bounds
            At the end of the optimisation result.value = preprocessor.inverse(result.value) will be applied

        :return: a structure optimisation result
        """
        # pylint: disable=too-many-branches
        outcome = argparse.Namespace()

        if isinstance(descriptor, fingerprinting.MomentInvariantsDescriptor):
            # Special case when the function being optimised is a descriptor
            if isinstance(target, base_moments.Moments):
                calc = descriptor.process[:-1]
            elif isinstance(target, np.ndarray):
                calc = descriptor.process
            else:
                raise TypeError(f'Unsupported type {target.__class__.__name__}')

            if preprocess:
                preprocessor = descriptor.preprocess

            bounds = bounds or descriptor.get_bounds(initial.num_atoms)

        else:
            calc = descriptor
            preprocess = False
            if preprocessor is not None:
                preprocess = True

        if preprocess:
            # Perform preprocessing
            initial = preprocessor(initial)
            if mask is not None:
                mask = preprocessor(mask)
            if bounds is not None:
                bounds = tuple(map(preprocessor, bounds))

        # Deal with saving of trajectory
        save_traj_fn = None
        if get_trajectory:
            outcome.traj = []
            save_traj_fn = functools.partial(self._save_trajectory, outcome.traj, preprocessor if preprocess else None)
        else:
            outcome.traj = None

        result = self._least_squares_optimiser.optimise_target(
            func=calc,
            initial=initial,
            target=target,
            mask=mask,
            jacobian=jacobian,
            bounds=bounds,
            max_func_evals=max_func_evals,
            x_tol=x_tol,
            cost_tol=cost_tol,
            grad_tol=grad_tol,
            callback=save_traj_fn,
            verbose=verbose
        )

        outcome.__dict__.update(result._asdict())

        if preprocess:
            # 'Un-preprocess' the output (map atomic species into integers)
            outcome.value = preprocessor.inverse(result.value)

        return StructureOptimisationResult(**outcome.__dict__)

    @staticmethod
    def _save_trajectory(trajectory: List, preprocessor, state: atomic.AtomsCollection, _value: np.ndarray, jacobian):
        if jacobian is not None:
            # Don't bother saving on Jacobian calls
            return

        if preprocessor is not None:
            state = preprocessor.inverse(state)

        trajectory.append(asetools.milad2ase(state))



# -*- coding: utf-8 -*-
from typing import Optional, Tuple, List

import numpy as np

from . import atomic
from . import base_moments
from . import invariants
from . import functions
from . import zernike

__all__ = 'MomentInvariantsDescriptor', 'descriptor', 'Fingerprinter', 'fingerprinter'


class MomentInvariantsDescriptor(functions.Function):
    """Class that is responsible for producing fingerprints form atomic environments"""
    scaler = None

    def __init__(
        self,
        feature_mapper: atomic.FeatureMapper,
        moments_calculator: base_moments.MomentsCalculator,
        invs: invariants.MomentInvariants,
        cutoff: float = None,
        scale: bool = True,
        species_mapper: atomic.MapNumbers = None,
        apply_cutoff=True,
        smooth_cutoff=False
    ):
        super().__init__()

        # PREPROCESSING
        self._preprocess = functions.Chain()
        self._species_mapper = species_mapper
        if self._species_mapper is not None:
            self._preprocess.append(self._species_mapper)

        # Now the actual fingerprinting
        self._invariants = invs

        # PROCESSING
        # Create the actual fingerprinting process which is a chain of functions
        process = functions.Chain()
        if cutoff is not None:
            if apply_cutoff:
                process.append(atomic.ApplyCutoff(cutoff))

            if scale:
                # Rescale positions to be in the range |r| < 1, the typical domain of orthogonality
                self.scaler = atomic.ScalePositions(1. / cutoff)
                process.append(self.scaler)

        process.append(feature_mapper)
        if smooth_cutoff:
            process.append(functions.CosineCutoff(cutoff))

        process.append(moments_calculator)
        process.append(self._invariants)

        self._cutoff = cutoff
        self._process = process
        self._moments_calculator = moments_calculator
        # Combine the two steps in one calculator
        self._calculator = functions.Chain(self._preprocess, self._process)

    @property
    def invariants(self) -> invariants.MomentInvariants:
        return self._invariants

    @property
    def fingerprint_len(self) -> int:
        return len(self._invariants)

    @property
    def cutoff(self) -> Optional[float]:
        return self._cutoff

    @property
    def preprocess(self) -> functions.Chain:
        """Return the preprocessing function"""
        return self._preprocess

    @property
    def process(self) -> functions.Chain:
        """Return the processing function"""
        return self._process

    @property
    def moments_calculator(self) -> functions.Function:
        """Get the function to calculate moments.  Note this is _exclusive_ of preprocessing which must be done
        separately if required.  If you just want to calculate the moments (including preprocessing) then use
        get_moments() instead."""
        return self.process[:-1]

    @property
    def species(self) -> Optional[List[int]]:
        """Get the species (as integers) supported by this descriptor.  Returns None if there is no restriction"""
        if self._species_mapper is None:
            return None

        return self._species_mapper.numbers

    def get_moments(self, atoms: atomic.AtomsCollection, preprocess=True) -> base_moments.Moments:
        if preprocess:
            atoms = self.preprocess(atoms)
        return self.process[:-1](atoms)

    def evaluate(self, state: atomic.AtomsCollection, *, get_jacobian=False):
        result = self._calculator(state, get_jacobian)
        if get_jacobian:
            return result[0].real, result[1].real

        return result.real

    def fingerprint_and_derivatives(self, atoms: atomic.AtomsCollection) -> Tuple[np.ndarray, np.ndarray]:
        """
        This method computes the fingerprint for the pass atoms collection and the corresponding position
        derivatives.
        :param atoms: the atoms collection to fingerprint
        :return: a tuple containing the fingerprint and the Jacobian
        """
        num_atoms = atoms.num_atoms

        # First preprocess as this may not be fully differentiable (and hence have no Jacobian)
        preprocessed = self.preprocess(atoms)

        # Now perform the rest of the fingerprinting procedure which does have derivatives
        fingerprint, jacobian = self.process(preprocessed, jacobian=True)
        len_fingerprint = len(fingerprint)

        # Now extract the portion of the Jacobian that relates just to atomic positions and reshape
        derivatives = jacobian[:, :3 * num_atoms].real
        # Reshape to be (len_fingerprint, num_atoms, 3) so xyz are stored in separate dimension
        derivatives = derivatives.reshape((len_fingerprint, num_atoms, 3))
        # Now sum xyz
        derivatives = derivatives.sum(axis=1)

        return fingerprint.real, derivatives

    def atom_centred(self, atoms: atomic.AtomsCollection, idx: int, get_jacobian=False):
        new_centre = atoms.positions[idx]
        new_atoms = atomic.AtomsCollection(
            atoms.num_atoms, positions=atoms.positions - new_centre, numbers=atoms.numbers
        )
        return self.evaluate(new_atoms, get_jacobian=get_jacobian)

    def get_bounds(self, num_atoms: int = 1) -> Tuple[atomic.AtomsCollection, atomic.AtomsCollection]:
        """Get a tuple containing a min and max AtomCollections that correspond to the minimum and maximum positions
        and (optionally) species range depending on the settings of this fingerprint.  This can be useful for setting
        optimiser bounds.
        """
        cutoff = self._cutoff

        # Figure out if we have a species range
        species_range = None

        # Let's look inside the preprocessing step to see if we're mapping atomic numbers, in which case we can
        # use this to set bounds on the species
        results = self.preprocess.find_type(atomic.MapNumbers)
        if results:
            numbers = results[0][1].numbers
            species_range = min(numbers), max(numbers)

        lower = atomic.AtomsCollection(num_atoms)
        upper = atomic.AtomsCollection(num_atoms)
        lower.vector[:] = -np.inf
        upper.vector[:] = np.inf

        if species_range:
            lower.numbers = species_range[0]
            upper.numbers = species_range[1]

        if cutoff is not None:
            lower.positions = -cutoff  # pylint: disable=invalid-unary-operand-type
            upper.positions = cutoff

        return lower, upper


def descriptor(
    features: Optional[dict] = None,
    species: Optional[dict] = None,
    cutoff: float = None,
    scale=True,
    moments_calculator: base_moments.MomentsCalculator = None,
    invs: invariants.MomentInvariants = None,
    apply_cutoff=True,
    smooth_cutoff=False,
):
    """

    :param features:
    :param species: a dictionary that has the following form:
        {
            'map': {
                'numbers': Sequence[Number] - a sequence of atomic numbers to be mapped
                'range': Union[Number, Tuple[Number, Number]] - a number range to map species to
                'to': Union[int, str] - the feature value to map the numbers to e.g. 'WEIGHT' which is a class
                                            property of WeightedDelta
            }
        }
    :param cutoff:
    :param scale: if True scale the environments by a factor of 1 / cutoff to fit within typical orthogonality region
    :param moments_calculator: the moment calculator to use
    :param invs: the invariants to use
    :return:
    """
    # Set up the preprocessing
    species = species or {}
    species_map = species.get('map', {})
    if species_map:
        species_mapper = atomic.MapNumbers(species=species_map['numbers'], map_to=species_map['range'])
    else:
        species_mapper = None

    # Default to Zernike moments if not supplied
    invs = invs or invariants.read(invariants.COMPLEX_INVARIANTS)
    moments_calculator = moments_calculator or zernike.ZernikeMomentCalculator(invs.max_order)

    features = features or dict(type=functions.WeightedDelta, map_species_to=species_map.get('to', None))
    return MomentInvariantsDescriptor(
        feature_mapper=atomic.FeatureMapper(**features),
        moments_calculator=moments_calculator,
        invs=invs,
        cutoff=cutoff,
        scale=scale,
        species_mapper=species_mapper,
        apply_cutoff=apply_cutoff,
        smooth_cutoff=smooth_cutoff,
    )


# Aliases for backwards compatibility
Fingerprinter = MomentInvariantsDescriptor
fingerprinter = descriptor



# -*- coding: utf-8 -*-
from typing import Callable

import numpy as np


class Data:
    """Data used during an optimisation"""

    def __init__(self, use_jacobian, complex_input: bool, verbose=False, builder=None, callback: Callable = None):
        self.use_jacobian = use_jacobian
        self.complex_input = complex_input
        self.verbose = verbose
        self.builder = builder
        self.callback = callback
        self.iter = 0


def join_state(state: np.ndarray) -> np.ndarray:
    half_size = int(len(state) / 2)
    out_state = state[:half_size] + 1j * state[half_size:]
    return out_state


def split_state(state: np.ndarray) -> np.ndarray:
    """Split is a state vector that is (potentially) complex.  If it is the vector is double in rows to be
    (reals, imags), otherwise the vector is simply returned unaltered."""
    if np.iscomplexobj(state):
        if isinstance(state, np.ndarray):
            return np.concatenate((state.real, state.imag))
        return np.array((state.real, state.imag))

    return state


def split_jacobian(jacobian: np.ndarray, complex_inputs: bool, complex_outputs: bool) -> np.ndarray:
    """Assuming that the function is analytic split the Jacobian into parts corresponding to a combination of
    complex inputs/outputs.  The layout of the return Jacobian is:

    | d Re(f)   d Re(f) |
    | -------   ------- |
    | d Re(x)   d Im(x) |
    |                   |
    | d Im(f)   d Im(f) |
    | -------   ------- |
    | d Re(x)   d Im(x) |

    If the function has complex inputs and real outputs then only the first row is return.
    If the function has real inputs and complex outputs then only the first column is return.
    Otherwise the full Jacobian is returned.

    See the Cauchy-Riemann equations for more details:
    https://en.wikipedia.org/wiki/Cauchy%E2%80%93Riemann_equations
    """
    if not complex_inputs and not complex_outputs:
        return jacobian.real

    orig_size = jacobian.shape
    if len(orig_size) == 1:
        orig_size = (orig_size[0], 1)
    new_size = list(orig_size)
    if complex_outputs:
        new_size[0] *= 2
    if complex_inputs:
        new_size[1] *= 2

    jac = np.empty(new_size)
    # Input: Real, Output: Real
    jac[:orig_size[0], :orig_size[1]] = jacobian.real
    if complex_outputs:
        # Input: Real, Output: Imag
        jac[orig_size[0]:, :orig_size[1]] = jacobian.imag
    if complex_inputs:
        complex_part = jacobian * 1j
        # Input: Imag, Output: Real
        jac[:orig_size[0], orig_size[1]:] = complex_part.real
        if complex_outputs:
            # Input: Imag, Output: Imag
            jac[orig_size[0]:, orig_size[1]:] = complex_part.imag

    return jac



# -*- coding: utf-8 -*-
import numpy
from scipy.spatial.transform import Rotation

import milad
from milad import transform

# def center(positions: numpy.array):
#     """Centre the given positions by shifting the centre of mass to zero"""
#     positions -= positions.sum(axis=0) / len(positions)


def test_general(moment_invariants):
    # Settings
    scale = 1.
    sigma = scale * 0.4
    num_rotations = 0
    num_points = 4

    pos1 = 4 * numpy.random.rand(num_points, 3)

    pos1 = transform.center(pos1)

    pos1 *= scale

    invariants = milad.invariants.calc_moment_invariants(
        moment_invariants, pos1, sigma)

    for rot in range(1, num_rotations + 1):
        rot = Rotation.random()
        rotated = rot.apply(pos1)

        # Calculate the moment invariants
        rot_invariants = milad.invariants.calc_moment_invariants(
            moment_invariants, rotated, sigma)

        assert numpy.testing.assert_array_almost_equal(invariants,
                                                       rot_invariants)



# -*- coding: utf-8 -*-
from typing import Union, Any, Callable, Dict, Tuple, List

import ase.neighborlist
import miniball
import numpy as np

import dscribe.utils.geometry

import milad
from milad import atomic
from milad import fingerprinting
from . import distances
from . import envs
from . import fingerprints


def calculate_fingerprint(atoms: ase.Atoms, calculator: fingerprints.FingerprintCalculator):
    num_atoms = len(atoms)
    nlist = ase.neighborlist.NeighborList(
        cutoffs=[0.5 * calculator.cutoff] * num_atoms, self_interaction=False, bothways=True
    )
    nlist.update(atoms)

    fps = []

    for idx, pos in enumerate(atoms.positions):
        indices, offsets = nlist.get_neighbors(idx)

        positions = []
        for i, offset in zip(indices, offsets):
            positions.append(atoms.positions[i] + np.dot(offset, atoms.get_cell()))

        # positions.append(pos)

        fps.append(calculator.calculate_neighbours(pos, positions))

    return np.array(fps)


def calculate_fingerprints_dscribe(atoms: ase.Atoms, calculator: fingerprints.FingerprintCalculator):
    positions = atoms.positions
    cutoff = calculator.cutoff
    cutoff_sq = cutoff * cutoff

    system = dscribe.utils.geometry.get_extended_system(atoms, cutoff, return_cell_indices=False)

    fps = []
    for i, pos in enumerate(positions):
        neighbours = []
        for j, neighbour in enumerate(system.positions):
            if i == j:
                continue

            dr = pos - neighbour
            dr_sq = np.dot(dr, dr)
            if dr_sq < cutoff_sq:
                neighbours.append(neighbour)

        fps.append(calculator.calculate_neighbours(pos, neighbours))

    return np.array(fps)


def _create_params_dict(species: list, value: Union[Any, dict]) -> dict:
    if isinstance(value, dict):
        return value

    return {specie: value for specie in species}


class MiladFingerprint:

    def __init__(
        self,
        species: list,
        invariants=None,
        sigmas=0.4,
        weights=1.0,
        cutoffs: Union[float, dict] = 6.,
        cutoff_fn=None,
        normalise=True,
        split_specie_pairs=False
    ):
        """
        :param species: a list of the species to consider
        :param invariants:
        :param sigmas:
        :param weights:
        :param cutoffs:
        :param cutoff_fn:
        :param normalise:
        :param split_specie_pairs: return a fingerprint that is a matrix containing specie pairs
        """
        self._species = species
        self._invariants = invariants or milad.invariants.read_invariants()
        self._sigmas = _create_params_dict(species, sigmas)
        self._weights = _create_params_dict(species, weights)
        self._cutoffs = _create_params_dict(species, cutoffs)
        self._cutoff_fns = _create_params_dict(species, cutoff_fn)
        self._normalise = normalise
        self._split_species = split_specie_pairs

    def set_params(
        self,
        specie: str,
        sigma: float = None,
        weight: float = None,
        cutoff: float = None,
        cutoff_fn: Union[float, Callable] = None
    ):
        """Set parameters for a particular specie"""
        if sigma is not None:
            self._sigmas[specie] = sigma
        if weight is not None:
            self._weights[specie] = weight
        if cutoff is not None:
            self._cutoffs[specie] = cutoff
        if cutoff_fn is not None:
            self._cutoff_fns[specie] = cutoff_fn

    def create(self, system: ase.Atoms, *args, **kwargs):
        positions = system.positions
        symbols = system.symbols

        fingerprint = np.zeros((len(positions), len(self._invariants)))

        if any(system.get_pbc()):
            dist_calculator = distances.UnitCellDistanceCalculator(system.cell.array)
        else:
            dist_calculator = distances.AperiodicDistanceCalculator()

        # Calculate the fingerprints for each atom centre one at a time
        for i, (pos, symbol) in enumerate(zip(positions, symbols)):
            env = envs.SmoothGaussianEnvironment(pos, self._cutoffs[symbol], self._cutoff_fns[symbol])

            # Add ourselves
            env.add_gaussian(pos, sigma=self._sigmas[symbol], weight=self._weights[symbol])
            cutoff = self._cutoffs[symbol]

            # Visit all neighbours
            for j, neighbour_pos in enumerate(positions):
                neighbour_symbol = symbols[j]
                vecs = dist_calculator.get_vecs_between(pos, neighbour_pos, cutoff=cutoff, self_interation=i != j)

                for vec in vecs:
                    env.add_gaussian(
                        pos + vec, sigma=self._sigmas[neighbour_symbol], weight=self._weights[neighbour_symbol]
                    )

            fingerprint[i, :] = env.calc_moment_invariants(self._invariants, normalise=self._normalise)

        return fingerprint

    def create_old(self, system: ase.Atoms, *args, **kwargs):
        positions = system.positions
        symbols = system.symbols

        fingerprint = np.zeros((len(positions), len(self._invariants)))

        nlist = ase.neighborlist.NeighborList(
            cutoffs=tuple(map(lambda sym: 0.5 * self._cutoffs[sym], symbols)), self_interaction=False, bothways=True
        )
        nlist.update(system)

        # Calculate the fingerprints for each atom centre one at a time
        for i, (pos, symbol) in enumerate(zip(positions, symbols)):
            env = envs.SmoothGaussianEnvironment(pos, self._cutoffs[symbol], self._cutoff_fns[symbol])

            # Add ourselves
            env.add_gaussian(pos, sigma=self._sigmas[symbol], weight=self._weights[symbol])

            # Visit all neighbours
            indices, offsets = nlist.get_neighbors(i)
            for j, offset in zip(indices, offsets):
                neighbour_pos = positions[j] + np.dot(offset, system.get_cell())
                neighbour_symbol = symbols[j]
                env.add_gaussian(
                    neighbour_pos, sigma=self._sigmas[neighbour_symbol], weight=self._weights[neighbour_symbol]
                )

            fingerprint[i, :] = env.calc_moment_invariants(self._invariants, normalise=self._normalise)

        return fingerprint

    def create_single_old(self, system: ase.Atoms, index):
        positions = system.positions
        symbols = system.symbols

        # Calculate the fingerprints for each atom centre one at a time
        pos = positions[index]
        symbol = symbols[index]

        if self._split_species:
            neighbour_types = [[specie] for specie in self._species]
        else:
            neighbour_types = [self._species]

        n_types = len(neighbour_types)
        fingerprint = np.zeros((n_types, len(self._invariants)))

        # Create the environments
        environs = []

        for _ in range(n_types):
            env = envs.SmoothGaussianEnvironment(pos, self._cutoffs[symbol], self._cutoff_fns[symbol])
            # Add ourselves
            env.add_gaussian(pos, sigma=self._sigmas[symbol], weight=self._weights[symbol])
            environs.append(env)

        # Create the neighbour list
        nlist = ase.neighborlist.NeighborList(
            cutoffs=tuple(map(lambda sym: 0.5 * self._cutoffs[sym], symbols)), self_interaction=False, bothways=True
        )
        nlist.update(system)

        # Go through all the neighbours and add Gaussians on each
        indices, offsets = nlist.get_neighbors(index)
        for i, offset in zip(indices, offsets):
            neighbour_pos = positions[i] + np.dot(offset, system.get_cell())
            neighbour_symbol = symbols[i]

            for type_idx, types in enumerate(neighbour_types):
                if neighbour_symbol in types:
                    environs[type_idx].add_gaussian(
                        neighbour_pos, sigma=self._sigmas[neighbour_symbol], weight=self._weights[neighbour_symbol]
                    )

        # Finally compute all the fingerprints
        for type_idx, env in enumerate(environs):
            fingerprint[type_idx] = env.calc_moment_invariants(self._invariants, normalise=self._normalise)

        if len(fingerprint) == 1:
            return fingerprint[0]

        return fingerprint

    def create_single(self, system: ase.Atoms, index):
        # pylint: disable=too-many-locals

        positions = system.positions
        symbols = system.symbols

        # Calculate the fingerprints for each atom centre one at a time
        pos = positions[index]
        symbol = symbols[index]
        cutoff = self._cutoffs[symbol]

        if self._split_species:
            neighbour_types = [[specie] for specie in self._species]
        else:
            neighbour_types = [self._species]

        n_types = len(neighbour_types)
        fingerprint = np.zeros((n_types, len(self._invariants)))

        # Create the environments
        environs = []

        for _ in range(n_types):
            env = envs.SmoothGaussianEnvironment(pos, self._cutoffs[symbol], self._cutoff_fns[symbol])
            # Add ourselves
            env.add_gaussian(pos, sigma=self._sigmas[symbol], weight=self._weights[symbol])
            environs.append(env)

        if any(system.get_pbc()):
            dist_calculator = distances.UnitCellDistanceCalculator(system.cell.array)
        else:
            dist_calculator = distances.AperiodicDistanceCalculator()

        for i, neighbour_pos in enumerate(positions):
            neighbour_symbol = symbols[i]
            vecs = dist_calculator.get_vecs_between(pos, neighbour_pos, cutoff=cutoff, self_interation=i != index)
            for type_idx, types in enumerate(neighbour_types):
                if neighbour_symbol in types:
                    for vec in vecs:
                        environs[type_idx].add_gaussian(
                            pos + vec, sigma=self._sigmas[neighbour_symbol], weight=self._weights[neighbour_symbol]
                        )

        # Finally compute all the fingerprints
        for type_idx, env in enumerate(environs):
            fingerprint[type_idx] = env.calc_moment_invariants(self._invariants, normalise=self._normalise)

        if len(fingerprint) == 1:
            return fingerprint[0]

        return fingerprint


def extract_environments(system: ase.Atoms, atom_centered=True, cutoff=5., yield_indices=False):
    """Given an ase.Atoms this will extract atomic environments and yield them as new Atoms objects

    The central atom will always be at position 0 and the rest (if any) will follow.

    :param system: the atoms object to extract environments from
    :param atom_centered: if True will centre the new environments on
    :param cutoff: a radial cuttoff for defining each environment
    :param yield_indices: if True will yield a tuple (idx, Atoms) where idx is the global index of the central atom
    """
    # pylint: disable=too-many-locals

    positions = system.positions
    symbols = system.symbols

    # Create a distance calculator
    if any(system.get_pbc()):
        dist_calculator = distances.UnitCellDistanceCalculator(system.cell.array)
    else:
        dist_calculator = distances.AperiodicDistanceCalculator()

    # Go over each atom and create the environment
    for i, (central_pos, central_symbol) in enumerate(zip(positions, symbols)):
        env_positions = [central_pos if not atom_centered else np.zeros(3)]
        env_symbols = [central_symbol]

        # Visit all neighbours
        for j, neighbour_pos in enumerate(positions):
            neighbour_symbol = symbols[j]
            vecs = dist_calculator.get_vecs_between(central_pos, neighbour_pos, cutoff=cutoff, self_interation=i != j)

            # Get all vectors to that neighbour (could be more than one if periodic)
            for vec in vecs:
                # Add the position
                env_positions.append(vec if atom_centered else vec + central_pos)
                # ...and the symbol
                env_symbols.append(neighbour_symbol)

        # Finally yield the environment
        if yield_indices:
            yield i, ase.Atoms(positions=env_positions, symbols=env_symbols)
        else:
            yield ase.Atoms(positions=env_positions, symbols=env_symbols)


class AseFingerprintsCalculator:
    """Convenience class for using ASE atoms objects with generic fingerprinting methods"""

    def __init__(self, fingerprinter: fingerprinting.MomentInvariantsDescriptor):
        self._fingerprinter = fingerprinter

    @property
    def fingerprinter(self):
        return self._fingerprinter

    def evaluate(self, atoms: ase.Atoms, get_jacobian=False):
        return self._fingerprinter(ase2milad(atoms), get_jacobian)

    def fingerprint_and_derivatives(self, atoms: ase.Atoms) -> Tuple[np.ndarray, np.ndarray]:
        return self._fingerprinter.fingerprint_and_derivatives(ase2milad(atoms))

    def __call__(self, atoms: ase.Atoms, get_jacobian=False):
        return self.evaluate(atoms, get_jacobian)


def ase2milad(atoms: ase.Atoms):
    return atomic.AtomsCollection(len(atoms), positions=atoms.positions, numbers=atoms.numbers)


def milad2ase(atoms: atomic.AtomsCollection) -> ase.Atoms:
    return ase.Atoms(positions=atoms.positions, numbers=atoms.numbers)


def prepare_molecule(*molecules: ase.Atoms, offset=0.) -> float:
    """This will bring the centroid of each molecule to be coincident with the origin and return
    the maximum radius found from any of the molecules"""
    max_radius_sq = 0.
    offset_vector = offset * np.array([1., 0., 0.])
    for molecule in molecules:
        centre, _radius = miniball.get_bounding_ball(molecule.positions)
        centroid = centre + offset_vector
        # Set the new positions
        molecule.set_positions(molecule.positions - centroid)
        new_positions = molecule.positions
        # Get the maximum radius squared
        max_dist_sq = max(np.dot(pos, pos) for pos in new_positions)
        max_radius_sq = max(max_radius_sq, max_dist_sq)

    return max_radius_sq**0.5



# -*- coding: utf-8 -*-
import abc
import copy
import logging
import numbers
from typing import List, Union, Tuple, Callable, Any, Iterable, Type, Optional

import numpy as np
import scipy

_LOGGER = logging.getLogger(__name__)


class State(metaclass=abc.ABCMeta):
    """A state vector representing an input to a function"""

    @property
    def dtype(self):
        return self.vector.dtype

    @property
    @abc.abstractmethod
    def vector(self) -> np.array:
        """Get the state vector as a numpy array"""

    @property
    def size(self):
        return self.array.size

    @property
    def array(self) -> np.array:
        """Get the state as a numpy array"""
        return self.vector

    def __len__(self) -> int:
        """Get the length of this state vector"""
        return len(self.vector)

    @property
    def length(self) -> int:
        """Get the length of this state vector"""
        return len(self.vector)

    def get_builder(self, mask=None) -> Optional['Function']:  # pylint: disable=no-self-use, disable=unused-argument
        """Optionally returns a builder that will convert a one dimensional array to an object of this type.
        Useful for optimisation or other algorithms that can't/don't want to deal with the internal structure of this
        object"""
        return None


StateLike = Union[np.ndarray, State]


class PlainState(State):
    __slots__ = ('_array',)

    def __init__(self, length, dtype=None):
        self._array = np.zeros(length, dtype=dtype)

    @property
    def vector(self) -> np.array:
        return self._array

    @property
    def array(self) -> np.array:
        """Get the array corresponding to this state"""
        return self._array

    @array.setter
    def array(self, arr: np.ndarray):
        self._array[:] = arr

    @property
    def dtype(self):
        return self._array.dtype


class Feature(State):
    LENGTH = None

    def __init__(self, dtype=float):
        super().__init__()
        if self.LENGTH is None:
            raise RuntimeError('Feature length not set, please set the LENGTH class attribute')
        self._vector = np.zeros(self.LENGTH, dtype=dtype)

    @property
    @abc.abstractmethod
    def pos(self) -> np.ndarray:
        """The position of this feature"""

    @property
    def vector(self):
        return self._vector

    @property
    def array(self):
        return self._vector

    @array.setter
    def array(self, arr: np.ndarray):
        self._vector[:] = arr

    @vector.setter
    def vector(self, value: np.array):
        if not value.size == self._vector.size:
            raise ValueError("Size mismatch, expected '{}', got '{}'".format(self._vector.size, value.size))
        self._vector[:] = value[:]

    def __add__(self, other: 'Feature') -> 'Features':
        if not isinstance(other, Feature):
            raise TypeError("Don't know how to add '{}' to a feature".format(other.__class__.__name__))
        return Features(self, other)


class Features(State):
    """An collection of features"""

    def __init__(self, *feature: Feature):
        self._vector = np.empty(0)
        self._features: List[Feature] = []
        for entry in feature:
            self.add(entry)

    @property
    def features(self) -> List[Feature]:
        """Access the list of features"""
        return self._features

    @property
    def vector(self):
        return self._vector

    @property
    def array(self):
        return self._vector

    @array.setter
    def array(self, arr):
        self._vector[:] = arr

    def feature_range(self, feature: Feature) -> Optional[range]:
        """Given a feature in this set of features, get the corresponding slice of indices in the feature vector"""
        try:
            idx = self._features.index(feature)
        except IndexError:
            return None

        start = sum([len(entry) for entry in self._features[:idx]])
        stop = start + len(feature)
        return range(start, stop)

    def add(self, feature: Feature):
        self._features.append(feature)
        self._vector = np.append(self._vector, feature.vector)
        # Reassign all the arrays
        idx = 0
        for entry in self._features:
            entry.array = self._vector[idx:idx + len(entry)]
            idx += entry.length

    def __add__(self, other: Union[Feature, 'Features']) -> 'Features':
        if not isinstance(other, (Feature, Features)):
            raise TypeError("Don't know how to add a '{}' to an environment".format(other.__class__.__name__))
        return Features(*self.features, other)


class WeightedDelta(Feature):
    """A Dirac delta function with a position and a weight (the total integral) stored as
    [x, y, z, weight]
    """
    # Indexing helpers
    X = 0
    Y = 1
    Z = 2
    WEIGHT = 3
    LENGTH = 4

    def __init__(self, pos: np.array, weight=1.0):
        super().__init__(dtype=pos.dtype)
        self.pos = pos
        self.weight = weight

    def __repr__(self) -> str:
        return f'Delta(pos={self.pos}, weight={self.weight})'

    @property
    def pos(self) -> np.array:
        return self.vector[:3]

    @pos.setter
    def pos(self, value):
        self.vector[:3] = value

    @property
    def weight(self):
        return self.vector[3]

    @weight.setter
    def weight(self, value):
        self.vector[3] = value


class WeightedGaussian(Feature):
    """A 3 dimensional gaussian stored as [x, y, z, sigma, weight]"""
    # Indexing helpers
    X = 0
    Y = 1
    Z = 2
    SIGMA = 3
    WEIGHT = 4
    LENGTH = 5

    def __init__(self, pos: np.array, sigma: float = 1., weight: float = 1.):
        super().__init__(dtype=pos.dtype)
        self.pos = pos
        self.sigma = sigma
        self.weight = weight

    def __repr__(self) -> str:
        return f'Gaussian(pos={self.pos}, sigma={self.sigma}, weight={self.weight})'

    @property
    def pos(self) -> np.array:
        return self.vector[:3]

    @pos.setter
    def pos(self, value):
        self.vector[:3] = value

    @property
    def sigma(self) -> float:
        return self.vector[self.SIGMA]

    @sigma.setter
    def sigma(self, value):
        self.vector[self.SIGMA] = value

    @property
    def weight(self) -> float:
        return self.vector[self.WEIGHT]

    @weight.setter
    def weight(self, value):
        self.vector[self.WEIGHT] = value


# region Functions


class Function(metaclass=abc.ABCMeta):
    input_type = (State, np.ndarray)
    output_type = (State, np.ndarray)
    supports_jacobian = False

    def __init__(self):
        self._callbacks = set()

    @property
    def inverse(self) -> Optional['Function']:
        """A function may optionally provide an inverse in which case it would be returned by this property"""
        return None

    def add_callback(self, func: Callable):
        if func in self._callbacks:
            raise ValueError("'{}' is already registered".format(func))
        self._callbacks.add(func)

    def remove_callback(self, func: Callable[[StateLike, StateLike, np.ndarray], None]):
        self._callbacks.remove(func)

    def __call__(self, state: State, jacobian=False):
        name = self.__class__.__name__

        if self.input_type is not None:
            self._check_input_type(state, self.input_type)

        try:
            input_vec = get_bare(state)
            if np.isnan(input_vec).any():
                raise ValueError(
                    f'{name} evaluate was passed an {state.__class__.__name__} input with a NaN value: {input_vec}'
                )
        except TypeError:
            pass  # The types stored in the array don't support isnan ufunc

        result = self.evaluate(state, get_jacobian=jacobian)
        if result is None:
            raise RuntimeError(f'{name} produced None output')

        if jacobian:
            if not isinstance(result, tuple):
                raise RuntimeError(f"{name}.evaluate didn't return Jacobian despite being asked to")
            val, jac = result[0], result[1]

            if not jac.dtype == object and np.isnan(jac).any():
                raise ValueError(f'{name}.evaluate produced a Jacobian with a NaN entry')
        else:
            val = result
            jac = None

        try:
            if np.isnan(get_bare(val)).any():
                raise ValueError(f'{name}.evaluate produced a result with a NaN entry')
        except TypeError:
            pass  # The types stored in the array don't support isnan ufunc

        for func in self._callbacks:
            if jacobian:
                func(state, val, jac)
            else:
                func(state, val, jacobian=None)

        return result

    @abc.abstractmethod
    def evaluate(self, state, *, get_jacobian=False):
        """Evaluate the function with the passed input

        :param state: the state that is the input to the function
        :param get_jacobian: if True returns a tuple [value, jacobian_mtx] else returns value
        """

    @classmethod
    def _check_input_type(cls, value: Any, allowed_types: Union[Type, Iterable[Type]]):
        cls._check_type(value, allowed_types, 'Unsupported input type')

    @classmethod
    def _check_output_type(cls, value: Any, allowed_types: Union[Type, Iterable[Type]]):
        cls._check_type(value, allowed_types, 'Unsupported output type')

    @classmethod
    def _check_type(cls, value: Any, allowed_types: Union[Type, Iterable[Type]], msg: str):
        if not isinstance(value, allowed_types):
            raise TypeError(
                f'{cls.__name__}: {msg}, '
                f"expected '{allowed_types}', "
                f"got '{value.__class__.__name__}'"
            )


class FromVectorBuilder(Function):
    """A default builder that can be used for state types that have a from_vector class method"""

    class Inverse(Function):

        def evaluate(self, state, *, get_jacobian=False):
            out = state.vector
            if get_jacobian:
                return out, np.eye(len(out))

            return out

    def __init__(self, state_type: Type, kwargs=None):
        super().__init__()
        self._state_type = state_type
        self._kwargs = kwargs or {}

    def evaluate(
        # pylint: disable=arguments-differ
        self,
        state_vec: np.ndarray,
        *,
        get_jacobian=False
    ):
        out = self._state_type.from_vector(state_vec, **self._kwargs)
        if get_jacobian:
            return out, np.eye(len(state_vec))

        return out

    @property
    def inverse(self) -> Optional['Function']:
        return FromVectorBuilder.Inverse()


class Identity(Function):
    """Function that just returns what it is passed"""

    def evaluate(self, state: StateLike, *, get_jacobian=False):
        """Evaluate the function with the passed input"""
        if get_jacobian:
            return state, np.eye(len(state))

        return state

    @property
    def inverse(self) -> Optional['Function']:
        return self


class Chain(Function):
    """A function that is a chain of other functions."""

    def __init__(self, *functions):
        super().__init__()
        for func in functions:
            if not isinstance(func, Function):
                raise TypeError(f'Expected Function, got {func.__class__.__name__}')
        self._functions: List[Function] = list(functions)

    def __len__(self):
        return len(self._functions)

    def __getitem__(self, item) -> Function:
        if isinstance(item, slice):
            return Chain(*self._functions[item])

        return self._functions[item]

    def __repr__(self):
        return f'Chain({self._functions})'

    def copy(self) -> 'Chain':
        """Create a shallow copy"""
        return Chain(self._functions)

    def find_type(self, query: type) -> List[Tuple[int, Function]]:
        """Look for functions in this chain of the given type.  Returns a list of Tuple[index, function]
        where index is the index in the chain and function is the matching function instance
        """
        found = []
        for idx, func in enumerate(self):
            if isinstance(func, query):
                found.append((idx, func))
        return found

    @property
    def input_type(self) -> Optional[type]:
        if not self._functions:
            return None

        return self._functions[0].input_type

    @property
    def output_type(self) -> Optional[type]:
        if not self._functions:
            return None

        return self._functions[-1].output_type

    @property
    def supports_jacobian(self):
        if not self._functions:
            return True

        return all(entry.supports_jacobian for entry in self._functions)

    @property
    def inverse(self) -> Optional['Function']:
        inverse_chain = Chain()
        for func in reversed(self._functions):
            inverse = func.inverse
            if inverse is None:
                # Can't build inverse chain as at least one function doesn't have an inverse
                return None
            inverse_chain.append(inverse)

        return inverse_chain

    def append(self, function: Function):
        if not isinstance(function, Function):
            raise TypeError(f'Expected Function, got {function.__class__.__name__}')
        self._functions.append(function)

    def evaluate(self, state: StateLike, *, get_jacobian=False):
        if not self._functions:
            return Identity()(state, jacobian=get_jacobian)

        if len(self._functions) == 1:
            # Just pipe directly through
            return self._functions[-1](state, get_jacobian)

        current_in = state
        previous_jacobian = None

        # Do all but last
        for function in self._functions[:-1]:
            # Compute values for the current function
            current_out = function(current_in, get_jacobian)
            if get_jacobian:
                current_out, current_jacobian = current_out

                if previous_jacobian is None:
                    # This is the first time around, so no previous to multiply with
                    previous_jacobian = current_jacobian
                else:
                    # Apply chain rule and propagate jacobians
                    previous_jacobian = np.matmul(current_jacobian, previous_jacobian)

            # Make this output the input to the next function
            current_in = current_out

        # Do the last one
        function = self._functions[-1]

        current_out = function(current_in, get_jacobian)
        if get_jacobian:
            current_out, current_jacobian = current_out
            return current_out, np.matmul(current_jacobian, previous_jacobian)

        return current_out


class Residuals(Function):
    """Given some measurements (m) calculate the residuals from some data (d) as f = d - m """
    supports_jacobian = True

    def __init__(self, data: np.array):
        super().__init__()
        self._data = get_bare(data)

    def output_length(self, in_state: State) -> int:  # pylint: disable=unused-argument
        return len(self._data)

    def empty_jacobian(self, in_state: State) -> np.array:
        return np.empty((self.output_length(in_state), len(in_state)), dtype=complex)

    def evaluate(self, state: State, *, get_jacobian=False) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        out_vector = get_bare(state) - self._data
        if get_jacobian:
            return out_vector, np.identity(len(state), dtype=out_vector.dtype)

        return out_vector


class Native(Function):

    def __init__(self, func: Callable[[StateLike], StateLike], jac: Callable[[StateLike], np.ndarray]):
        super().__init__()
        self._fn = func
        self._jac = jac

    @property
    def support_jacobian(self):
        return self._jac is not None

    def evaluate(self, state: State, *, get_jacobian=False) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        if get_jacobian:
            return self._fn(state), self._jac(state)

        return self._fn(state)


class CosineCutoff(Function):
    """A function that takes a set of Features and scales their weights by applying a cosine cutoff in the form

    w' = w (cos(\pi * dr / r_\text{cut}) + 1.) / 2
    """
    supports_jacobian = False

    def __init__(self, cutoff: float):
        super().__init__()
        self._cutoff = cutoff

    def evaluate(self, in_features: Features, *, get_jacobian=False):  # pylint: disable=arguments-differ
        out_features = Features()
        r_cut_sq = self._cutoff**2

        jacobians = []
        for in_feature in in_features.features:
            out_feature = None
            if hasattr(in_feature, 'weight') and hasattr(in_feature, 'pos'):
                r_sq = np.dot(in_feature.pos, in_feature.pos)

                # Hack that makes it possible to test this function analytically
                if r_sq < r_cut_sq:
                    dr = r_sq**0.5  # pylint: disable=invalid-name
                    out_feature = copy.deepcopy(in_feature)
                    # Rescale the weight
                    scale = self.fn(dr)
                    out_feature.weight = in_feature.weight * scale

                    if get_jacobian:
                        # Calculate this portion of the Jacobian
                        jac_block = np.eye(len(in_feature))
                        dr_dxyz = out_feature.pos / dr if dr > 0. else np.zeros(3)
                        dwout_dr = in_feature.weight * self.deriv(dr)

                        jac_block[in_feature.WEIGHT, 0:3] = dwout_dr * dr_dxyz
                        jac_block[in_feature.WEIGHT, in_feature.WEIGHT] = scale

                        jacobians.append(jac_block)

            else:
                # Just pass on the same feature, no changes
                out_feature = in_feature

                if get_jacobian:
                    # Calculate this portion of the Jacobian
                    jacobians.append(np.eye(len(in_feature)))

            if out_feature is not None:
                out_features.add(out_feature)

        if get_jacobian:
            return out_features, scipy.linalg.block_diag(*jacobians)

        return out_features

    def fn(
        # pylint: disable=invalid-name
        self,
        dr
    ) -> float:
        return 0.5 * (np.cos(np.pi * dr / self._cutoff) + 1.)

    def deriv(
        # pylint: disable=invalid-name
        self,
        dr
    ) -> float:
        return -0.5 * np.pi / self._cutoff * np.sin(np.pi * dr / self._cutoff)

    @staticmethod
    def symbolic(
        # pylint: disable=invalid-name
        dr,
        weight,
        rcut
    ):
        import sympy as sp
        return weight * (sp.cos(sp.pi * dr / rcut) + 1) / 2


# endregion


def get_bare(state: Union[np.ndarray, State]) -> Union[np.array, numbers.Number]:
    """This will return either a numpy array (through state.vector) if it is a state class, otherwise it will just
    return what it is passed.  The idea is that what is returned should be able to be operated on using standard
    mathematical operations."""
    if isinstance(state, State):
        return state.vector

    return state


def copy_to(
    dest: np.ndarray,
    indices: np.ndarray,
    source: np.ndarray,
    source_indices=None,
):
    source_indices = source_indices or tuple(range(len(source)))
    for desc_idx, source_idx in zip(indices, source_indices):
        dest[desc_idx] = source[source_idx]


def nan_check(array: np.ndarray, msg=None):
    """Check that an array does not contain NaN values.  If it does this will raise a ValueError"""
    if not array.dtype == object and np.isnan(array).any():
        exc_msg = 'Found NaN value'
        if msg is not None:
            exc_msg += f', {msg}'
        raise ValueError(msg)



# -*- coding: utf-8 -*-
from rdkit import Chem

from . import atomic


def milad2rdkit(atoms: atomic.AtomsCollection) -> Chem.Mol:
    """Covnert a MILAD atoms collection to an rdkit molecule"""
    mol = Chem.RWMol()
    conf = Chem.Conformer(atoms.num_atoms)
    for idx in range(atoms.num_atoms):
        atom = Chem.Atom(int(atoms.numbers[idx]))
        mol.AddAtom(atom)
        conf.SetAtomPosition(idx, atoms.positions[idx])
    mol.AddConformer(conf)
    return mol



import numpy

def generate_vectors_on_sphere(num):
    """Generate `num` random unit vectors on a sphere"""
    vecs = numpy.empty((num, 3))
    for i in range(num):
        vec = numpy.random.rand(3)
        vec /= numpy.linalg.norm(vec, axis=0)
        vecs[i] = vec
    return vecs


# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt
import numpy
from scipy.spatial.transform import Rotation

import milad

# pylint: disable=too-many-locals


def generate_vectors_on_sphere(num):
    vecs = numpy.empty((num, 3))
    for i in range(num):
        vec = numpy.random.rand(3)
        vec /= numpy.linalg.norm(vec, axis=0)
        vecs[i] = vec
    return vecs


def test_env_simple_no_cutoff_fn(moment_invariants, request):
    # Settings
    numpy.random.seed(5)
    num_atoms = 6
    sigma = 0.5
    volume = 1.
    num_rotations = 10
    sphere_radius = 2.5
    normalise = True

    positions = sphere_radius * generate_vectors_on_sphere(num_atoms)

    # Populate the environment
    env = milad.play.SmoothGaussianEnvironment()
    for pos in positions:
        env.add_gaussian(pos, sigma=sigma, weight=volume)

    # The original
    series = [env.calc_moment_invariants(moment_invariants, normalise=normalise)]

    for _ in range(num_rotations):
        # Generated rotated positions
        rot = Rotation.random()
        rotated = rot.apply(positions)

        # Create the environment
        rotated_env = milad.play.SmoothGaussianEnvironment()
        rotated_env.add_gaussians(rotated, sigma=sigma, mass=volume)

        # Calculate the invariants
        series.append(rotated_env.calc_moment_invariants(moment_invariants, normalise=normalise))

    fig, axes = plt.subplots()
    milad.plot.plot_multiple_invariants(series, axes)
    fig.legend()
    fig.savefig('{}.pdf'.format(request.node.name))


def test_env_simple_cos_cutoff_fn(moment_invariants, request):
    num_atoms = 6
    sigma = 1.0
    num_rotations = 10
    pos_sphere = 2.5
    # moment_invariants = moment_invariants[:20]
    cutoff_params = milad.play.make_cutoff_params(milad.play.cos_cutoff, 6.0)

    positions = pos_sphere * generate_vectors_on_sphere(num_atoms)

    # Populate the environment
    env = milad.play.SmoothGaussianEnvironment(**cutoff_params)
    for pos in positions:
        env.add_gaussian(pos, sigma=sigma)

    # The original
    series = [env.calc_moment_invariants(moment_invariants, normalise=True)]

    for _ in range(num_rotations):
        # Generated rotated positions
        rot = Rotation.random()
        rotated = rot.apply(positions)

        # Create the environment
        rotated_env = milad.play.SmoothGaussianEnvironment(**cutoff_params)
        _results = rotated_env.add_gaussians(rotated, sigma)

        # Calculate the invariants
        series.append(rotated_env.calc_moment_invariants(moment_invariants, normalise=True))

    fig, axes = plt.subplots()
    milad.plot.plot_multiple_invariants(series, axes)
    fig.legend()
    fig.savefig('{}.pdf'.format(request.node.name))


def test_atom_entering(moment_invariants, request):
    """Test how the invariants change as an atom enters the cutoff"""
    num_atoms = 6
    cutoff = 6.
    sigma = 1.0

    start_x = 6.1
    delta = 0.05
    num_steps = 7

    weight = 200.
    normalise = True
    # moment_invariants = moment_invariants[:20]
    cutoff_params = milad.play.make_cutoff_params(milad.play.cos_cutoff, cutoff)

    # The second atom just outside the cutoff zone
    positions = cutoff * numpy.random.rand(num_atoms, 3)
    target = num_atoms - 1
    positions[target] = (start_x, 0., 0.)

    # The original
    series = []

    for i in range(num_steps):
        x_pos = start_x - i * delta
        positions[target] = numpy.array((x_pos, 0., 0.))

        # Create the environment
        env = milad.play.SmoothGaussianEnvironment(**cutoff_params)
        # Add all the environment atoms
        for pos in positions[:target]:
            env.add_gaussian(pos, sigma=sigma, weight=weight)

        result = env.add_gaussian(positions[target], sigma=sigma, weight=weight)
        if x_pos > 6.:
            assert result is False

        # Calculate the invariants
        series.append(env.calc_moment_invariants(moment_invariants, normalise=normalise))

    diffs = [series[idx] - series[0] for idx in range(1, num_steps)]
    labels = [f'x={start_x - i * delta:.2f}' for i in range(1, num_steps)]

    fig, axes = plt.subplots()
    axes.set_title('Atom entering environment, $r_c=6$')
    # axes.set_yscale('log')
    axes.set_ylabel(f'Difference from x={start_x:.2f}')
    milad.plot.plot_multiple_invariants(diffs, axes, labels)
    fig.legend()
    fig.savefig('{}.pdf'.format(request.node.name))


def test_asymmetric_distribution(moment_invariants, request):
    """Test how the invariants change as an atom enters the cutoff"""
    sigma = 1.0
    normalise = True
    # moment_invariants = moment_invariants[:10]

    # The second atom just outside the cutoff zone
    pos = numpy.array((-3., 0., 0.))

    # Create the environment
    env = milad.play.SmoothGaussianEnvironment()
    env.add_gaussian(pos, sigma=sigma)

    series = []

    # Calculate the invariants
    series.append(env.calc_moment_invariants(moment_invariants, normalise=normalise))

    fig, axes = plt.subplots()
    milad.plot.plot_multiple_invariants(series, axes)
    fig.legend()
    fig.savefig('{}.pdf'.format(request.node.name))


def test_fingerprint(moment_invariants, request):
    num_atoms = 5
    num_rotations = 10
    scale = 4.
    num_invariants = len(moment_invariants)

    positions = scale * numpy.random.rand(num_atoms, 3)
    calculator = milad.play.FingerprintCalculator(
        invariants=moment_invariants, sigma=1., cutoff=6., cutoff_function='cos', normalise=True
    )

    orig_fp = calculator.calculate(positions)
    zeros = numpy.zeros((len(orig_fp)))

    series = [orig_fp]
    for _ in range(num_rotations):
        # Generated rotated positions
        rot = Rotation.random()
        rotated = rot.apply(positions)

        # Calculate the fingerprint
        fprint = calculator.calculate(rotated)
        series.append(fprint)

        delta = fprint - orig_fp
        numpy.testing.assert_almost_equal(delta, zeros)

    fig, axes = plt.subplots()
    axes.set_title('5 random atoms 6x6x6 cube, $\\sigma=1, r_{cut} = 6.$')
    milad.plot.plot_multiple_invariants(series, axes, labels=['$R_{{{}}}$'.format(idx) for idx in range(len(series))])
    axes.set_xlabel('Total fingerprint ({} atoms x {} invariants)'.format(num_atoms, num_invariants))
    fig.legend()
    fig.savefig(f'{request.node.name}.pdf')



# -*- coding: utf-8 -*-
"""Module for analytic operations"""
import itertools
from typing import Tuple, Union

import sympy


def create_array(indexable, shape: Union[int, Tuple]) -> sympy.MutableDenseNDimArray:
    """Given an indexed object this will create an array of the specified shape containing the
    indexable object indexed at each position"""
    if isinstance(shape, int):
        shape = (shape,)

    components = sympy.MutableDenseNDimArray.zeros(*shape)

    ranges = [range(entry) for entry in shape]
    for indices in itertools.product(*ranges):
        components[indices] = indexable[indices]

    return components



# -*- coding: utf-8 -*-
from .least_squares import *
from .moments_optimiser import *  # pylint: disable=undefined-variable
from .structure_optimiser import *  # pylint: disable=undefined-variable

__all__ = least_squares.__all__ + moments_optimiser.__all__ + structure_optimiser.__all__  # pylint: disable=undefined-variable
